"""Handles postprocessing of compiled ASTs for the Viv DSL."""

__all__ = ["postprocess_combined_ast"]

from viv_compiler import config, errors, external_types, internal_types, utils, validation
from ._inheritance import handle_inheritance
from ._role_dependencies import build_role_dependency_forests, build_condition_groups


def postprocess_combined_ast(combined_ast: internal_types.CombinedAST) -> None:
    """Postprocess the given combined AST by inserting higher-order metadata and performing other manipulations.

    Args:
        combined_ast: An abstract syntax tree produced by the Visitor class, integrating the
            respective ASTs of any included files.

    Returns:
        Nothing. Mutates the combined AST in place.
    """
    # Handle inheritance between action definitions
    handle_inheritance(combined_ast=combined_ast)
    # For each action selector that implicitly passes through an initiator role, create a (virtual)
    # initiator role for the selector (if needed) and populate the precast bindings for all
    # associated candidates. This step must precede construction of the role-dependency forests.
    _populate_initiator_pass_through_action_selectors(combined_ast=combined_ast)
    # Resolve any positional bindings in trope fits
    _resolve_positional_bindings(combined_ast=combined_ast)
    # Wrap applicable fields with their associated role references
    _wrap_expressions_with_role_references(combined_ast=combined_ast)
    # Mark initiator roles as `precast`
    _mark_initiator_roles_as_precast(combined_ast=combined_ast)
    # Set the `initiator` field in the action definitions
    _attribute_initiator_role(combined_ast=combined_ast)
    # Attribute to casting-pool directives whether they are cachable. This must come
    # after `_attribute_initiator_role()` has been called.
    _attribute_casting_pool_cachability_values(combined_ast=combined_ast)
    # For each construct definition, construct a dependency forest that will structure
    # role casting during targeting of that construct.
    build_role_dependency_forests(combined_ast=combined_ast)
    # Finally, convert the `conditions` field into a mapping from role name to (only)
    # the conditions that must be evaluated to cast that role.
    build_condition_groups(combined_ast=combined_ast)


def _populate_initiator_pass_through_action_selectors(combined_ast: internal_types.CombinedAST) -> None:
    """For each selector that implicitly passes through an initiator role, create a (virtual) initiator
    role for the selector (if needed) and populate the precast bindings for all associated candidates.

    If an author elides the bindings on a selector candidate, the initiator at hand (for targeting of
    the selector itself) will be passed on as the prospective initiator of the candidate. If an author
    also elides the `roles` field in the selector, we will create a virtual initiator role that the
    selector can pass through to its candidates as it targets them in turn.

    Args:
        combined_ast: A combined AST as produced by inheritance processing.

    Returns:
        None. Mutates the construct definitions in place.
    """
    # Before we start, prevalidate all selectors
    validation.prevalidate_action_selectors(combined_ast=combined_ast)
    # Create virtual initiator roles, as needed
    for selector_definition in combined_ast["actionSelectors"]:
        if not any(utils.is_initiator_role(role_definition=role) for role in selector_definition['roles'].values()):
            virtual_initiator_role = external_types.RoleDefinition(
                name=config.SELECTOR_VIRTUAL_INITIATOR_ROLE_NAME,
                entityType=external_types.RoleEntityType.CHARACTER,
                min=1,
                max=1,
                parent=None,
                children=[],
                pool=None,
                chance=None,
                mean=None,
                sd=None,
                participationMode=external_types.RoleParticipationMode.INITIATOR,
                anywhere=False,
                precast=False,
                spawn=False,
                spawnFunction=None,
                renames=None
            )
            selector_definition["roles"][virtual_initiator_role['name']] = virtual_initiator_role
    # Populate precast bindings for all candidates that were specified using the initiator pass-through pattern
    for selector_definition in combined_ast["actionSelectors"]:
        selector_initiator_role = next(
            role for role in selector_definition['roles'].values() if utils.is_initiator_role(role_definition=role)
        )
        selector_initiator_role_name = selector_initiator_role['name']
        for candidate in selector_definition["candidates"]:
            if candidate["bindings"]["roles"]:
                continue
            if candidate["isSelector"]:
                candidate_definition = next(
                    selector for selector in combined_ast['actionSelectors'] if selector['name'] == candidate['name']
                )
            else:
                candidate_definition = next(
                    action for action in combined_ast['actions'] if action['name'] == candidate['name']
                )
            candidate_initiator_role = next(
                role for role in candidate_definition['roles'].values()
                if utils.is_initiator_role(role_definition=role)
            )
            candidate_initiator_role_name = candidate_initiator_role["name"]
            candidate["bindings"]["roles"][candidate_initiator_role_name] = external_types.EntityReference(
                type=external_types.ExpressionDiscriminator.ENTITY_REFERENCE,
                value=external_types.ReferenceValue(
                    anchor=selector_initiator_role_name,
                    path=[],
                    local=False,
                    group=False,
                    failSafe=False
                ),
                source=None,
                negated=False
            )


def _resolve_positional_bindings(combined_ast: internal_types.CombinedAST) -> None:
    """Resolve any positional bindings in trope fits.

    This function only errors when it can't do its work of resolving positional bindings, which occurs
    with the given issues stated below. During the actual validation phase, we will work more thoroughly
    to catch things like missing bindings.

    Args:
        combined_ast: A combined AST as produced by inheritance processing.

    Returns:
        None. Mutates the construct definitions in place.

    Raises:
        VivCompileError: A trope fit targets an undefined trope.
        VivCompileError: A trope fit has too many positional bindings.
    """
    for construct_definition in utils.get_all_construct_definitions(combined_ast=combined_ast):
        all_trope_fit_full_expressions: list[external_types.TropeFit] = utils.get_all_full_expressions_of_type(
            expression_type=external_types.ExpressionDiscriminator.TROPE_FIT,
            ast_chunk=construct_definition
        )
        for trope_fit_full_expression in all_trope_fit_full_expressions:
            trope_fit_expression: external_types.TropeFitValue = trope_fit_full_expression['value']
            trope_fit_source = trope_fit_full_expression['source']
            # Retrieve the trope definition. If it targets an undefined trope, throw an error.
            try:
                trope_definition = next(
                    trope for trope in combined_ast['tropes'] if trope['name'] == trope_fit_expression['tropeName']
                )
            except StopIteration:
                raise errors.VivCompileError(
                    msg=(
                        f"{utils.CONSTRUCT_LABEL[construct_definition['type']]} '{construct_definition['name']}' "
                        f"references undefined trope: '{trope_fit_expression['tropeName']}'"
                    ),
                    source=trope_fit_source
                )
            # Resolve any positional bindings
            trope_fit_binding_role_names = list(trope_fit_expression['bindings']['roles'])
            for bound_role_name in trope_fit_binding_role_names:
                # The Viv grammar does not allow mixing of positional and named bindings, so we can
                # immediately move on from any trope fit with a named binding.
                if not bound_role_name.startswith(config.POSITIONAL_BINDING_TRANSIENT_PREFIX):
                    break
                # If we have an extra binding, throw an error now
                role_position = int(bound_role_name[len(config.POSITIONAL_BINDING_TRANSIENT_PREFIX):])
                if role_position >= len(trope_definition['roles']):  # Positional bindings are 0-indexed
                    n_roles = len(trope_definition['roles'])
                    n_roles_precast = len(trope_fit_binding_role_names)
                    raise errors.VivCompileError(
                        msg=(
                            f"{utils.CONSTRUCT_LABEL[construct_definition['type']]} "
                            f"'{construct_definition['name']}' invokes trope "
                            f"'{trope_fit_expression['tropeName']}' with too many positional "
                            f"bindings (expected at most {n_roles}, got {n_roles_precast})"
                        ),
                        source=trope_fit_source
                    )
                # Otherwise, resolve the positional binding by attributing the role name associated
                # with that slot. Note that dictionary keys have been ordered since Python 3.7.
                resolved_role_name = list(trope_definition['roles'].keys())[role_position]
                trope_fit_expression['bindings']['roles'][resolved_role_name] = (
                    trope_fit_expression['bindings']['roles'][bound_role_name]
                )
                del trope_fit_expression['bindings']['roles'][bound_role_name]


def _wrap_expressions_with_role_references(combined_ast: internal_types.CombinedAST) -> None:
    """Wrap applicable fields with their role references.

    Args:
        combined_ast: A combined AST as produced by inheritance processing.

    Returns:
        None. Mutates the construct definitions in place.
    """
    # For all constructs, we'll wrap conditions
    for construct_definition in utils.get_all_construct_definitions(combined_ast=combined_ast):
        condition_wrappers = []
        for condition in construct_definition["_conditions_raw"]:
            wrapper = external_types.WrappedExpression(
                body=condition,
                references=utils.get_all_referenced_roles(ast_chunk=condition)
            )
            condition_wrappers.append(wrapper)
        construct_definition["_conditions_raw"] = condition_wrappers
    # For actions, we'll also wrap effects and reactions, producing the final schema-compliant field values
    for action_definition in combined_ast["actions"]:
        effect_wrappers = []
        for effect in action_definition["_effects_raw"]:
            wrapper = external_types.WrappedExpression(
                body=effect,
                references=utils.get_all_referenced_roles(ast_chunk=effect)
            )
            effect_wrappers.append(wrapper)
        action_definition["effects"] = effect_wrappers
        reaction_wrappers = []
        for reaction in action_definition["_reactions_raw"]:
            wrapper = external_types.WrappedExpression(
                body=reaction,
                references=utils.get_all_referenced_roles(ast_chunk=reaction)
            )
            reaction_wrappers.append(wrapper)
        action_definition["reactions"] = reaction_wrappers


def _attribute_initiator_role(combined_ast: internal_types.CombinedAST) -> None:
    """Modify the given action and selector definitions in place, to attribute to each an initiator role.

    Args:
        combined_ast: A combined AST as produced by inheritance processing.

    Returns:
        None (modifies the action and selector definitions in place).

    Raises:
        VivCompileError: If an action definition or selector definition does not define an initiator role.
    """
    validation.prevalidate_initiator_roles(combined_ast=combined_ast)  # Ensures there's always exactly one initiator
    for construct_definition in utils.get_all_construct_definitions(combined_ast=combined_ast):
        if construct_definition['type'] in config.CONSTRUCT_TYPES_USING_INITIATOR_ROLES:
            initiator_role: external_types.RoleDefinition = next(
                role for role in construct_definition['roles'].values()
                if utils.is_initiator_role(role_definition=role)
            )
            construct_definition['initiator'] = initiator_role['name']


def _mark_initiator_roles_as_precast(combined_ast: internal_types.CombinedAST) -> None:
    """Modify the given construct definitions in place to mark initiator roles as `precast`.

    Args:
        combined_ast: A combined AST as produced by inheritance processing.

    Returns:
        None. Mutates the construct definitions in place.
    """
    for construct_definition in utils.get_all_construct_definitions(combined_ast=combined_ast):
        for role_definition in construct_definition['roles'].values():
            role_definition: external_types.RoleDefinition
            # If this is an initiator role, explicitly attribute the 'precast' label
            if utils.is_initiator_role(role_definition=role_definition):
                role_definition["precast"] = True


def _attribute_casting_pool_cachability_values(combined_ast: internal_types.CombinedAST) -> None:
    """Modify the given action definitions in place to mark casting-pool declarations as cachable/uncachable.

    Currently only action roles can have cachable casting-pool declarations. These are attributed
    when a pool declaration does not reference a non-initiator role.

    Args:
        combined_ast: A combined AST as produced by inheritance processing.

    Returns:
        None. Mutates the construct definitions in place.
    """
    for action_definition in combined_ast["actions"]:
        for role_definition in action_definition['roles'].values():
            if not role_definition['pool']:
                continue
            role_definition['pool']['uncachable'] = False
            all_pool_references = utils.get_all_referenced_roles(role_definition['pool'])
            if any(role for role in all_pool_references if role != action_definition['initiator']):
                role_definition['pool']['uncachable'] = True



"""Module that creates the finalized compiled content bundle.

This module takes in a postprocessed combined AST produced by the Viv postprocessor, and uses
the construct definitions contained in that AST to construct a compiled content bundle, which
will also contain metadata produced via another module.
"""

from .bundler import create_compiled_content_bundle



"""Custom error classes containing detailed information about a compilation issue."""

import textwrap
from pathlib import Path

from arpeggio import NoMatch

from viv_compiler import external_types


class VivCompileError(Exception):
    """Error raised when Viv compilation fails.

    Attributes:
        msg: A brief message explaining the issue.
        file_path: Absolute path to the offending source file, if applicable, else None.
        line: Line number at the start of the offending source, if applicable, else None.
        column: Column number at the start of the offending source, if applicable, else None.
        end_line: Line number at the end of the offending source, if applicable, else None.
        end_column: Column number at the end of the offending source, if applicable, else None.
        code: The offending source code snippet, if applicable, else None.
    """

    def __init__(self, msg: str, source: external_types.SourceAnnotations | None = None):
        """Initialize a VivCompileError object.

        Args:
            msg: A brief message explaining the issue.
            source: If applicable, source annotations identifying the Viv source code that caused the issue.
        """
        self.msg = msg
        # If source annotations are available, populate the location fields
        if source:
            self.file_path = source["filePath"]
            self.line = source["line"]
            self.column = source["column"]
            self.end_line = source["endLine"]
            self.end_column = source["endColumn"]
            self.code = source["code"]
        else:
            self.file_path = None
            self.line = None
            self.column = None
            self.end_line = None
            self.end_column = None
            self.code = None
        # Use our full string representation in the call to the parent's constructor, which sets `self.msg`
        super().__init__(str(self))

    def __str__(self) -> str:
        """Return full string representation."""
        if self.code:
            normalized = textwrap.dedent(self.code).strip()
            indented = textwrap.indent(normalized, "  ")
            return (
                f"{self.msg}\n\n"
                f"{self.file_path!r}:{self.line}:{self.column}\n\n"
                f"{indented}"
            )
        return self.msg

    def __repr__(self) -> str:
        """Return compact string representation."""
        if self.file_path:
            return f"{type(self).__name__}({self.msg!r}, {self.file_path!r}:{self.line}:{self.column})"
        return f"{type(self).__name__}({self.msg!r})"


class VivParseError(VivCompileError):
    """Error raised when Arpeggio fails to parse Viv source code.

    Attributes:
        msg: A brief message explaining the issue.
        file_path: Absolute path to the offending source file, if applicable, else None.
        original: The original error thrown by Arpeggio.
        original_formatted: The original error thrown by Arpeggio, in a friendlier format.
        line: Line number where the parse error occurred, if applicable, else None.
        column: Column number where the parse error occurred, if applicable, else None.
    """

    def __init__(self, original: Exception, file_path: Path):
        """Initialize a VivParseError object.

        Args:
            original: The original error thrown by Arpeggio.
            file_path: Path for the file that could not be parsed. This may be an included file.
        """
        self.file_path = file_path
        self.original = original
        self.original_formatted = self._format_arpeggio_error(error=original, file_path=file_path)
        msg = f"Source file could not be parsed"
        super().__init__(msg=msg)  # Must come before the lines below
        if isinstance(original, NoMatch):
            self.line = original.line
            self.column = original.col
        else:
            self.line = None
            self.column = None
        self.msg = str(self)

    def __str__(self) -> str:
        """Return full string representation."""
        return f"{self.msg}:\n\n{self.original_formatted}"

    def __repr__(self) -> str:
        """Return compact string representation."""
        return f"{type(self).__name__}({self.file_path!r})"

    @staticmethod
    def _format_arpeggio_error(error: Exception, file_path: Path) -> str:
        """Returns the given Arpeggio error message in a friendlier format.

        Args:
            error: An Arpeggio parsing error. If it's something else, that's fine.
            file_path: Path for the file that could not be parsed. This may be an included file.

        Returns:
            A formatted error message.
        """
        if not isinstance(error, NoMatch):
            return str(error)
        error.eval_attrs()
        snippet = error.context.strip()
        expected_tokens = error.message.replace('Expected ', '', 1).replace(" or ", ", ").replace("'", '')
        formatted_error = (
            f"- File: {file_path}"
            f"\n- Position: line {error.line}, col {error.col}"
            f"\n- Context (failed at *): ...{snippet.replace('*', '[*]')}..."
            f"\n- Viable next tokens: {expected_tokens}"
        )
        return formatted_error



"""Version identifiers for concerns pertinent to the Viv compiler."""

__all__ = ["__version__", "__grammar_version__", "__schema_version__"]

import json
from typing import Final
from importlib import resources
from importlib.metadata import version, PackageNotFoundError


def _read_compiler_version() -> str:
    """Return the current compiler version.

    Note: If the package is being run directly from source, a dummy version is returned.

    Returns:
        The current compiler version.
    """
    try:
        return version("viv-compiler")
    except PackageNotFoundError:
        # If the compiler is being run directly from source, return a dummy version
        return "0.0.0+local"


def _read_grammar_version() -> str:
    """Return the current grammar version stored in the grammar file.

    Returns:
        The current grammar version.

    Raises:
        RuntimeError: There is no version number in the grammar file.
    """
    text = resources.files("viv_compiler.grammar").joinpath("viv.peg").read_text(encoding="utf-8")
    for line in text.splitlines():
        if line.startswith("// VERSION:"):
            return line.split(":", 1)[1].strip()
    raise RuntimeError("Grammar version not found") from None


def _read_schema_version() -> str:
    """Return the current schema version stored in the canonical compiler schema file.

    Returns:
        The current schema version.

    Raises:
        RuntimeError: There is no version number in the schema file.
    """
    text = resources.files("viv_compiler.schemas").joinpath("compiler.schema.json").read_text(encoding="utf-8")
    try:
        return json.loads(text)["version"]
    except KeyError:
        raise RuntimeError("Schema version not found") from None


# Version for the Viv compiler
__version__: Final[str] = _read_compiler_version()

# Version for the Viv DSL grammar (editor plugins assume compatibility)
__grammar_version__: Final[str] = _read_grammar_version()

# Version for the schema for compiled content bundles (runtimes enforce compatibility)
__schema_version__: Final[str] = _read_schema_version()



"""Mixin class containing visitor methods associated with action definitions."""

__all__ = ["VisitorMixinActions"]

from copy import deepcopy
from typing import Any, Final

from arpeggio import NonTerminal, PTNodeVisitor

from viv_compiler import config, errors, external_types, internal_types, sentinels, utils, validation
from ._statements import VisitorMixinStatements


class VisitorMixinActions(PTNodeVisitor):
    """A visitor mixin for Viv actions."""

    # The name of the transient `_join_directives` field
    JOIN_DIRECTIVES_FIELD_NAME: Final = "_join_directives"
    # A default value for the `tags` field, used when the author elides it
    TAGS_DEFAULT = external_types.ListField(
        type=external_types.ExpressionDiscriminator.LIST,
        value=[],
        source=None
    )
    # A default value for the `importance` field, used when the author elides it
    IMPORTANCE_DEFAULT: Final = external_types.FloatField(
        type=external_types.ExpressionDiscriminator.FLOAT,
        value=config.DEFAULT_IMPORTANCE_SCORE,
        source=None
    )
    # A default value for the `default` subfield of the `associations` field, used when the author elides it
    ASSOCIATIONS_DEFAULT_FIELD_DEFAULT: Final = external_types.ListField(
        type=external_types.ExpressionDiscriminator.LIST,
        value=[],
        source=None
    )
    # A default value for the `default` subfield of the `saliences` field, used when the author elides it
    SALIENCES_DEFAULT_FIELD_DEFAULT: Final = external_types.FloatField(
        type=external_types.ExpressionDiscriminator.FLOAT,
        value=config.DEFAULT_SALIENCE_SCORE,
        source=None
    )

    @staticmethod
    def visit_action(
        node: NonTerminal,
        children: list[Any]
    ) -> internal_types.IntermediateActionDefinition | internal_types.IntermediateChildActionDefinition:
        """Visit an <action> node."""
        # Populate the fields that were authored
        accumulated_fields = {}
        action_header = children[0]
        action_body = children[1] if len(children) > 1 else {}  # Bare renaming of a parent action
        accumulated_fields.update(action_header)
        for field_name, field_value in action_body.items():
            if field_name == "roles":
                validation.prevalidate_role_names(
                    construct_type=external_types.ConstructDiscriminator.ACTION,
                    construct_name=accumulated_fields['name'],
                    role_definitions=field_value,
                    source=utils.derive_source_annotations(node=node)
                )
                accumulated_fields["roles"] = {role["name"]: role for role in field_value}
            else:
                accumulated_fields[field_name] = field_value
        # Ensure `roles` is present, unless it's a child action
        if "roles" not in action_body:
            if "parent" not in accumulated_fields:
                error_message = (
                    f"Action '{accumulated_fields['name']}' is missing 'roles' field "
                    f"(required for actions that do not inherit from a parent action)"
                )
                raise errors.VivCompileError(error_message, source=utils.derive_source_annotations(node=node))
        # Fill out the minimal intermediate structure, seeding with default values as applicable
        if "parent" in accumulated_fields:
            # If it's a child action, we won't seed defaults for any heritable fields, because
            # such defaults would override what would otherwise be inherited from the parent.
            intermediate_action_definition = internal_types.IntermediateChildActionDefinition(
                type=external_types.ConstructDiscriminator.ACTION,
                name=accumulated_fields["name"],
                _template=accumulated_fields.get("_template", False),
                reserved=accumulated_fields.get("reserved", False),
                parent=accumulated_fields["parent"],
                _join_directives=accumulated_fields.get(VisitorMixinActions.JOIN_DIRECTIVES_FIELD_NAME, [])
            )
            # Add in any fields that were specified for the child. This will duplicate some of
            # the work we just carried out in manually setting required fields, but we'll keep
            # the above so that we're sure to at least set all required fields.
            intermediate_action_definition.update(accumulated_fields)
        else:
            intermediate_action_definition = internal_types.IntermediateActionDefinition(
                type=external_types.ConstructDiscriminator.ACTION,
                name=accumulated_fields["name"],
                _template=accumulated_fields.get("_template", False),
                reserved=accumulated_fields.get("reserved", False),
                parent=accumulated_fields.get("parent", None),
                roles=accumulated_fields["roles"],
                importance=accumulated_fields.get("importance", deepcopy(VisitorMixinActions.IMPORTANCE_DEFAULT)),
                tags=accumulated_fields.get("tags", deepcopy(VisitorMixinActions.TAGS_DEFAULT)),
                gloss=accumulated_fields.get("gloss", None),
                report=accumulated_fields.get("report", None),
                _conditions_raw=accumulated_fields.get("_conditions_raw", []),
                scratch=accumulated_fields.get("scratch", []),
                _effects_raw=accumulated_fields.get("_effects_raw", []),
                _reactions_raw=accumulated_fields.get("_reactions_raw", []),
                saliences=accumulated_fields.get("saliences", external_types.Saliences(
                    default=deepcopy(VisitorMixinActions.SALIENCES_DEFAULT_FIELD_DEFAULT),
                    roles={},
                    custom=[],
                    variable=None
                )),
                associations=accumulated_fields.get("associations", external_types.Associations(
                    default=deepcopy(VisitorMixinActions.ASSOCIATIONS_DEFAULT_FIELD_DEFAULT),
                    roles={},
                    custom=[],
                    variable=None
                )),
                embargoes=accumulated_fields.get("embargoes", [])
            )
        return intermediate_action_definition

    @staticmethod
    def visit_action_header(_, children: list[Any]) -> dict[str, Any]:
        """Visit an <action_header> node."""
        action_header = {}
        for child in children:
            if isinstance(child, str):
                action_header['name'] = child
            else:
                action_header.update(child)
        return action_header

    @staticmethod
    def visit_reserved_construct_marker(_, __) -> dict[str, bool]:
        """Visit a <reserved_construct_marker> node."""
        return {"reserved": True}

    @staticmethod
    def visit_template_action_marker(_, __) -> dict[str, bool]:
        """Visit a <template_action_marker> node."""
        return {"_template": True}

    @staticmethod
    def visit_parent_action_declaration(
        _, children: list[external_types.ActionName]
    ) -> dict[str, external_types.ActionName]:
        """Visit a <parent_action_declaration> node."""
        return {"parent": children[0]}

    @staticmethod
    def visit_action_body(_, children: list[Any]) -> dict[str, Any]:
        """Visit an <action_body> node."""
        accumulated_fields = {}
        for child in children:
            for field_name, field_value in child.items():
                if field_name == VisitorMixinActions.JOIN_DIRECTIVES_FIELD_NAME:
                    accumulated_fields.setdefault(VisitorMixinActions.JOIN_DIRECTIVES_FIELD_NAME, [])
                    accumulated_fields[VisitorMixinActions.JOIN_DIRECTIVES_FIELD_NAME].append(field_value)
                else:
                    accumulated_fields[field_name] = field_value
        return accumulated_fields

    @staticmethod
    def visit_action_gloss(
        _, children: list[external_types.TemplateStringField | external_types.StringField]
    ) -> dict[str, external_types.TemplateStringField | external_types.StringField]:
        """Visit an <action_gloss> node."""
        gloss = children[0]
        return {"gloss": gloss}

    @staticmethod
    def visit_action_report(
        _, children: list[external_types.TemplateStringField | external_types.StringField]
    ) -> dict[str, external_types.TemplateStringField | external_types.StringField]:
        """Visit an <action_report> node."""
        return {"report": children[0]}

    @staticmethod
    def visit_action_tags(_, children: list[Any]) -> dict[str, Any]:
        """Visit an <action_tags> node."""
        if children[0] is sentinels.CHILD_JOIN_OPERATOR:
            accumulated_fields = {
                "tags": children[1],
                VisitorMixinActions.JOIN_DIRECTIVES_FIELD_NAME: sentinels.CHILD_JOIN_DIRECTIVE_TAGS
            }
        else:
            accumulated_fields = {"tags": children[0]}
        return accumulated_fields

    @staticmethod
    def visit_action_importance(
        _, children: list[external_types.IntField | external_types.FloatField | external_types.Enum]
    ) -> dict[str, external_types.IntField | external_types.FloatField | external_types.Enum]:
        """Visit an <action_importance> node."""
        return {"importance": children[0]}

    @staticmethod
    def visit_action_roles(_, children: list[Any]) -> dict[str, Any]:
        """Visit an <action_roles> node."""
        if children[0] is sentinels.CHILD_JOIN_OPERATOR:
            accumulated_fields = {
                "roles": children[1:],
                VisitorMixinActions.JOIN_DIRECTIVES_FIELD_NAME: sentinels.CHILD_JOIN_DIRECTIVE_ROLES
            }
        else:
            accumulated_fields = {"roles": children}
        return accumulated_fields

    @staticmethod
    def visit_action_conditions(_, children: list[Any]) -> dict[str, Any]:
        """Visit an <action_conditions> node."""
        if children[0] is sentinels.CHILD_JOIN_OPERATOR:
            accumulated_fields = {
                "_conditions_raw": children[1],
                VisitorMixinActions.JOIN_DIRECTIVES_FIELD_NAME: sentinels.CHILD_JOIN_DIRECTIVE_CONDITIONS
            }
        else:
            accumulated_fields = {"_conditions_raw": children[0]}
        return accumulated_fields

    @staticmethod
    def visit_action_scratch(_, children: list[Any]) -> dict[str, Any]:
        """Visit an <action_scratch> node."""
        if children[0] is sentinels.CHILD_JOIN_OPERATOR:
            accumulated_fields = {
                "scratch": children[1],
                VisitorMixinActions.JOIN_DIRECTIVES_FIELD_NAME: sentinels.CHILD_JOIN_DIRECTIVE_SCRATCH
            }
        else:
            accumulated_fields = {"scratch": children[0]}
        return accumulated_fields

    @staticmethod
    def visit_action_effects(_, children: list[Any]) -> dict[str, Any]:
        """Visit an <action_effects> node."""
        if children[0] is sentinels.CHILD_JOIN_OPERATOR:
            accumulated_fields = {
                "_effects_raw": children[1],
                VisitorMixinActions.JOIN_DIRECTIVES_FIELD_NAME: sentinels.CHILD_JOIN_DIRECTIVE_EFFECTS
            }
        else:
            accumulated_fields = {"_effects_raw": children[0]}
        return accumulated_fields

    @staticmethod
    def visit_action_reactions(_, children: list[Any]) -> dict[str, Any]:
        """Visit an <action_reactions> node."""
        if children[0] is sentinels.CHILD_JOIN_OPERATOR:
            accumulated_fields = {
                "_reactions_raw": children[1:],
                VisitorMixinActions.JOIN_DIRECTIVES_FIELD_NAME: sentinels.CHILD_JOIN_DIRECTIVE_REACTIONS
            }
        else:
            accumulated_fields = {"_reactions_raw": children}
        return accumulated_fields

    @staticmethod
    def visit_action_saliences(_, children: list[external_types.Saliences]) -> dict[str, external_types.Saliences]:
        """Visit an <action_saliences> node."""
        return {"saliences": children[0]}

    @staticmethod
    def visit_saliences_body(_, children: list[Any]) -> external_types.Saliences:
        """Visit a <saliences_body> node."""
        accumulated_fields = {}
        for child in children:
            accumulated_fields.update(child)
        saliences = external_types.Saliences(
            default=accumulated_fields.get("default", deepcopy(VisitorMixinActions.SALIENCES_DEFAULT_FIELD_DEFAULT)),
            roles=accumulated_fields.get("roles", {}),
            custom=accumulated_fields.get("custom", []),
            variable=accumulated_fields.get("variable", None)
        )
        return saliences

    @staticmethod
    def visit_saliences_default(
        _, children: list[external_types.SalienceScoreExpression]
    ) -> dict[str, external_types.SalienceScoreExpression]:
        """Visit a <saliences_default> node."""
        return {"default": children[0]}

    @staticmethod
    def visit_saliences_roles(
        node: NonTerminal,
        children: list[Any]
    ) -> dict[str, dict[str, external_types.SalienceScoreExpression]]:
        """Visit a <saliences_roles> node."""
        role_entries = {}
        for role_entry in children:
            for role_name, expression in role_entry.items():  # It's just a single key–value pair, but easy to iterate
                if role_name in role_entries:
                    error_message = f"Duplicate entry detected in action saliences for role '{role_name}'"
                    raise errors.VivCompileError(error_message, source=utils.derive_source_annotations(node=node))
                role_entries[role_name] = expression
        return {"roles": role_entries}

    @staticmethod
    def visit_saliences_roles_entry(_, children: list[Any]) -> dict[str, external_types.SalienceScoreExpression]:
        """Visit a <saliences_roles_entry> node."""
        _sigil, name, expression = children
        return {name: expression}

    @staticmethod
    def visit_saliences_custom_field(_, children: list[Any]) -> dict[str, Any]:
        """Visit a <saliences_custom_field> node."""
        local_variable, *expressions = children
        accumulated_fields = {
            "custom": expressions,
            "variable": local_variable
        }
        return accumulated_fields

    @staticmethod
    def visit_action_associations(
        _, children: list[external_types.Associations]
    ) -> dict[str, external_types.Associations]:
        """Visit an <action_associations> node."""
        return {"associations": children[0]}

    @staticmethod
    def visit_associations_body(_, children: list[Any]) -> external_types.Associations:
        """Visit an <associations_body> node."""
        accumulated_fields = {}
        for child in children:
            accumulated_fields.update(child)
        associations = external_types.Associations(
            default=accumulated_fields.get("default", deepcopy(VisitorMixinActions.ASSOCIATIONS_DEFAULT_FIELD_DEFAULT)),
            roles=accumulated_fields.get("roles", {}),
            custom=accumulated_fields.get("custom", []),
            variable=accumulated_fields.get("variable", None)
        )
        return associations

    @staticmethod
    def visit_associations_default(_, children: list[external_types.ListField]) -> dict[str, external_types.ListField]:
        """Visit an <associations_default> node."""
        return {"default": children[0]}

    @staticmethod
    def visit_associations_roles(
        node: NonTerminal, children: list[Any]
    ) -> dict[str, dict[str, external_types.ListField]]:
        """Visit an <associations_roles> node."""
        role_entries = {}
        for role_entry in children:
            for role_name, expression in role_entry.items():  # It's just a single key–value pair, but easy to iterate
                if role_name in role_entries:
                    error_message = f"Duplicate entry detected in action associations for role '{role_name}'"
                    raise errors.VivCompileError(error_message, source=utils.derive_source_annotations(node=node))
                role_entries[role_name] = expression
        return {"roles": role_entries}

    @staticmethod
    def visit_associations_roles_entry(_, children: list[Any]) -> dict[str, external_types.ListField]:
        """Visit an <associations_roles_entry> node."""
        _sigil, name, expression = children
        return {name: expression}

    @staticmethod
    def visit_associations_custom_field(_, children: list[Any]) -> dict[str, Any]:
        """Visit an <associations_custom_field> node."""
        local_variable, *expressions = children
        accumulated_fields = {
            "custom": expressions,
            "variable": local_variable
        }
        return accumulated_fields

    @staticmethod
    def visit_associations_statement(
        _, children: list[list[external_types.Expression]]
    ) -> list[external_types.Expression]:
        """Visit an <associations_statement> node."""
        return children[0]

    @staticmethod
    def visit_associations_loop(node: NonTerminal, children: list[Any]) -> external_types.Loop:
        """Visit an <associations_loop> node."""
        return VisitorMixinStatements.visit_loop(node=node, children=children)

    @staticmethod
    def visit_associations_conditional(node: NonTerminal, children: list[Any]) -> external_types.Conditional:
        """Visit an <associations_conditional> node."""
        return VisitorMixinStatements.visit_conditional(node=node, children=children)

    @staticmethod
    def visit_associations_conditional_branches(
        _, children: list[external_types.ConditionalBranch]
    ) -> list[external_types.ConditionalBranch]:
        """Visit an <associations_conditional_branches> node."""
        return VisitorMixinStatements.visit_conditional_branches(_, children)

    @staticmethod
    def visit_associations_conditional_branch(_, children: list[Any]) -> external_types.ConditionalBranch:
        """Visit an <associations_conditional_branch> node."""
        return VisitorMixinStatements.visit_conditional_branch(_, children)

    @staticmethod
    def visit_associations_conditional_consequent(
        _, children: list[list[external_types.Expression]]
    ) -> list[external_types.Expression]:
        """Visit an <associations_conditional_consequent> node."""
        return VisitorMixinStatements.visit_consequent(_, children)

    @staticmethod
    def visit_associations_conditional_alternative(
        _, children: list[list[external_types.Expression]]
    ) -> list[external_types.Expression]:
        """Visit an <associations_conditional_alternative> node."""
        return VisitorMixinStatements.visit_alternative(_, children)

    @staticmethod
    def visit_associations_scoped_statements(
        _, children: list[external_types.Expression]
    ) -> list[external_types.Expression]:
        """Visit an <associations_scoped_statements> node."""
        return children

    @staticmethod
    def visit_action_embargoes(_, children: list[Any]) -> dict[str, Any]:
        """Visit an <action_embargoes> node."""
        if children[0] is sentinels.CHILD_JOIN_OPERATOR:
            accumulated_fields = {
                "embargoes": children[1:],
                VisitorMixinActions.JOIN_DIRECTIVES_FIELD_NAME: sentinels.CHILD_JOIN_DIRECTIVE_EMBARGOES
            }
        else:
            accumulated_fields = {"embargoes": children}
        return accumulated_fields

    @staticmethod
    def visit_embargo(_, children: list[external_types.EmbargoDeclaration]) -> external_types.EmbargoDeclaration:
        """Visit an <embargo> node."""
        return children[0]

    @staticmethod
    def visit_embargo_body(node: NonTerminal, children: list[Any]) -> external_types.EmbargoDeclaration:
        """Visit an <embargo_body> node."""
        accumulated_fields = {}
        for child in children:
            accumulated_fields.update(child)
        if "permanent" not in accumulated_fields and "period" not in accumulated_fields:
            raise errors.VivCompileError("Embargo is missing 'time' field (required)", source=utils.derive_source_annotations(node=node))
        embargo_declaration = external_types.EmbargoDeclaration(
            roles=accumulated_fields.get("roles", None),
            permanent=accumulated_fields.get("permanent", False),
            period=accumulated_fields.get("period", None),
            here=accumulated_fields.get("here", False),
        )
        return embargo_declaration

    @staticmethod
    def visit_embargo_roles(
        _, children: list[internal_types.IntermediateRoleReference]
    ) -> dict[str, list[external_types.RoleName]]:
        """Visit an <embargo_roles> node."""
        return {"roles": [child['_name'] for child in children]}

    @staticmethod
    def visit_embargo_time_period(_, children: list[Any]) -> dict[str, Any]:
        """Visit an <embargo_time_period> node."""
        raw_time_period = children[0]
        if raw_time_period is sentinels.PERMANENT_EMBARGO_MARKER:
            accumulated_fields = {"permanent": True, "period": None}
        else:
            time_period = external_types.TimeDelta(amount=raw_time_period["amount"], unit=raw_time_period["unit"])
            accumulated_fields = {"permanent": False, "period": time_period}
        return accumulated_fields

    @staticmethod
    def visit_embargo_time_period_forever(_, __) -> internal_types.Sentinel:
        """Visit an <embargo_time_period_forever> node."""
        return sentinels.PERMANENT_EMBARGO_MARKER

    @staticmethod
    def visit_embargo_location(_, children: list[internal_types.Sentinel]) -> dict[str, Any]:
        """Visit an <embargo_location> node."""
        return {"here": children[0] is sentinels.LOCATION_SPECIFIC_EMBARGO_MARKER}

    @staticmethod
    def visit_embargo_location_here(_, __) -> internal_types.Sentinel:
        """Visit an <embargo_location_here> node."""
        return sentinels.LOCATION_SPECIFIC_EMBARGO_MARKER

    @staticmethod
    def visit_embargo_location_anywhere(_, __) -> internal_types.Sentinel:
        """Visit an <embargo_location_anywhere> node."""
        return sentinels.LOCATION_AGNOSTIC_EMBARGO_MARKER

    @staticmethod
    def visit_child_join_operator(_, __) -> internal_types.Sentinel:
        """Visit a <child_join_operator> node."""
        return sentinels.CHILD_JOIN_OPERATOR



"""Internal types that are used by the compiler only.

The types defined here describe shapes whose life cycles begin and expire during
compilation. As such, they are not part of the public API for the compiler.
"""

from __future__ import annotations

from typing import Literal, NotRequired, Required, TypeAlias, TypedDict

from viv_compiler import external_types


class CombinedAST(TypedDict):
    """The merged AST after import resolution.

    Note: The `includes` field has been handled and is no longer present here.
    """
    # All intermediate actions across the entry file and its resolved imports
    actions: list[IntermediateActionDefinition | IntermediateChildActionDefinition]
    # All intermediate action-selector definitions across the entry file and its resolved imports
    actionSelectors: list[IntermediateActionSelectorDefinition]
    # All intermediate plan definitions across the entry file and its resolved imports
    plans: list[IntermediatePlanDefinition]
    # All intermediate plan-selector definitions across the entry file and its resolved imports
    planSelectors: list[IntermediatePlanSelectorDefinition]
    # All intermediate query definitions across the entry file and its resolved imports
    queries: list[IntermediateQueryDefinition]
    # All intermediate sifting-pattern definitions across the entry file and its resolved imports
    siftingPatterns: list[IntermediateSiftingPatternDefinition]
    # All intermediate trope definitions across the entry file and its resolved imports
    tropes: list[IntermediateTropeDefinition]


class AST(CombinedAST):
    """Visitor output for a single source file, prior to import resolution."""
    # A list of relative import paths, exactly as authored in the source file (using the
    # `include` operator). This transient field is deleted once inheritance has been handled.
    _includes: list[str]


class IntermediateConstructDefinitionBase(TypedDict, total=False):
    """A base shape that is shared across all intermediate construct definitions.

    A lone exception to this is intermediate child actions, which exist only before inheritance
    handling and have only a few required properties.
    """
    # A mapping from role name to role definition, for all roles defined for the construct
    roles: Required[dict[external_types.RoleName, external_types.RoleDefinition]]
    # A list containing all conditions on the construct (wrapping occurs midway through postprocessing)
    _conditions_raw: Required[list[external_types.Expression] | list[external_types.WrappedExpression]]


class IntermediateActionDefinition(IntermediateConstructDefinitionBase, total=False):
    """An intermediate action definition.

    If the action definition was originally a child action, this shape takes hold following inheritance.

    The shape captures the structure of action definitions throughout postprocessing.
    """
    # The `parent` field in its final schema-compliant shape, used as a discriminator distinguishing
    # an `IntermediateActionDefinition` from an `IntermediateChildActionDefinition`.
    parent: Required[None]
    # The discriminator for the action construct
    type: Required[external_types.ConstructDiscriminator.ACTION]
    # A list containing all effects defined for the action (pre-wrapping)
    _effects_raw: Required[list[external_types.Expression]]
    # A list containing all reactions declared for the action (pre-wrapping)
    _reactions_raw: Required[list[external_types.Reaction]]
    # Whether this is a transient 'template action' that is only intended to be inherited
    # by child actions, after which point it is deleted.
    _template: Required[bool]
    # (Assigned during postprocessing) The `initiator` field in its final schema-compliant shape
    initiator: external_types.RoleName
    # (Assigned during postprocessing) The `roleForestRoots` field in its final schema-compliant shape
    roleForestRoots: list[external_types.RoleName]
    # (Assigned during postprocessing) The `conditions` field in its final schema-compliant shape
    conditions: external_types.ConstructConditions
    # (Assigned during postprocessing) The `effects` field in its final schema-compliant shape
    effects: list[external_types.WrappedExpression]
    # (Assigned during postprocessing) The `reactions` field in its final schema-compliant shape
    reactions: list[external_types.WrappedExpression]
    # The `name` field in its final schema-compliant shape
    name: Required[external_types.ActionName]
    # The `reserved` field in its final schema-compliant shape
    reserved: Required[bool]
    # The `tags` field in its final schema-compliant shape
    tags: Required[external_types.ListField]
    # The `gloss` field in its final schema-compliant shape
    gloss: Required[external_types.TemplateStringField | None]
    # The `report` field in its final schema-compliant shape
    report: Required[external_types.TemplateStringField | None]
    # The `importance` field in its final schema-compliant shape
    importance: Required[external_types.IntField | external_types.FloatField | external_types.Enum]
    # The `scratch` field in its final schema-compliant shape
    scratch: Required[list[external_types.Expression]]
    # The `saliences` field in its final schema-compliant shape
    saliences: Required[external_types.Saliences]
    # The `associations` field in its final schema-compliant shape
    associations: Required[external_types.Associations]
    # The `embargoes` field in its final schema-compliant shape
    embargoes: Required[list[external_types.EmbargoDeclaration]]


class IntermediateChildActionDefinition(TypedDict, total=False):
    """An intermediate child action definition, prior to inheritance handling.

    The shape captures the structure of child action definitions prior to inheritance handling,
    when they have only a few required properties.
    """
    # The `parent` field in its final schema-compliant shape, used as a discriminator distinguishing
    # an `IntermediateActionDefinition` from an `IntermediateChildActionDefinition`.
    parent: Required[external_types.ActionName]
    # Directives each indicating whether to merge a given field value with that of a parent,
    # as opposed to overriding the parent field value entirely.
    _join_directives: Required[list[Sentinel]]
    # The discriminator for the action construct
    type: Required[external_types.ConstructDiscriminator.ACTION]
    # A list containing all effects defined for the action (pre-wrapping)
    _effects_raw: list[external_types.Expression]
    # A list containing all reactions declared for the action (pre-wrapping)
    _reactions_raw: list[external_types.Reaction]
    # Whether this is a transient 'template action' that is only intended to be inherited
    # by child actions, after which point it is deleted.
    _template: Required[bool]
    # The `name` field in its final schema-compliant shape
    name: Required[external_types.ActionName]
    # The `reserved` field in its final schema-compliant shape
    reserved: Required[bool]
    # The `tags` field in its final schema-compliant shape
    tags: external_types.ListField
    # The `gloss` field in its final schema-compliant shape
    gloss: external_types.TemplateStringField | None
    # The `report` field in its final schema-compliant shape
    report: external_types.TemplateStringField | None
    # The `importance` field in its final schema-compliant shape
    importance: external_types.IntField | external_types.FloatField | external_types.Enum
    # The `scratch` field in its final schema-compliant shape
    scratch: list[external_types.Expression]
    # The `saliences` field in its final schema-compliant shape
    saliences: external_types.Saliences
    # The `associations` field in its final schema-compliant shape
    associations: external_types.Associations
    # The `embargoes` field in its final schema-compliant shape
    embargoes: list[external_types.EmbargoDeclaration]


class IntermediatePlanDefinition(IntermediateConstructDefinitionBase, total=False):
    """An intermediate plan definition.

    The shape captures the structure of plan definitions throughout postprocessing.
    """
    # The discriminator for the plan construct
    type: Required[Literal[external_types.ConstructDiscriminator.PLAN]]
    # (Assigned during postprocessing) The `roleForestRoots` field in its final schema-compliant shape
    roleForestRoots: list[external_types.RoleName]
    # (Assigned during postprocessing) The `conditions` field in its final schema-compliant shape
    conditions: external_types.ConstructConditions
    # The `name` field in its final schema-compliant shape
    name: Required[external_types.PlanName]
    # The `phases` field in its final schema-compliant shape
    phases: Required[dict[external_types.PlanPhaseName, external_types.PlanPhase]]
    # The `initialPhase` field in its final schema-compliant shape
    initialPhase: Required[external_types.PlanPhaseName]


class IntermediateQueryDefinition(IntermediateConstructDefinitionBase, total=False):
    """An intermediate query definition.

    The shape captures the structure of query definitions throughout postprocessing.
    """
    # The discriminator for the query construct
    type: Required[Literal[external_types.ConstructDiscriminator.QUERY]]
    # (Assigned during postprocessing) The `roleForestRoots` field in its final schema-compliant shape
    roleForestRoots: list[external_types.RoleName]
    # (Assigned during postprocessing) The `conditions` field in its final schema-compliant shape
    conditions: external_types.ConstructConditions
    # The `name` field in its final schema-compliant shape
    name: Required[external_types.QueryName]
    # The `actionName` field in its final schema-compliant shape
    actionName: Required[list[external_types.SetPredicate] | None]
    # The `ancestors` field in its final schema-compliant shape
    ancestors: Required[list[external_types.SetPredicate] | None]
    # The `descendants` field in its final schema-compliant shape
    descendants: Required[list[external_types.SetPredicate] | None]
    # The `importance` field in its final schema-compliant shape
    importance: Required[external_types.QueryNumericRange | None]
    # The `tags` field in its final schema-compliant shape
    tags: Required[list[external_types.SetPredicate] | None]
    # The `salience` field in its final schema-compliant shape
    salience: Required[external_types.QueryNumericRange | None]
    # The `associations` field in its final schema-compliant shape
    associations: Required[list[external_types.SetPredicate] | None]
    # The `location` field in its final schema-compliant shape
    location: Required[list[external_types.SetPredicate] | None]
    # The `time` field in its final schema-compliant shape
    time: Required[list[external_types.TemporalConstraint] | None]
    # The `initiator` field in its final schema-compliant shape
    initiator: Required[list[external_types.SetPredicate] | None]
    # The `partners` field in its final schema-compliant shape
    partners: Required[list[external_types.SetPredicate] | None]
    # The `recipients` field in its final schema-compliant shape
    recipients: Required[list[external_types.SetPredicate] | None]
    # The `bystanders` field in its final schema-compliant shape
    bystanders: Required[list[external_types.SetPredicate] | None]
    # The `active` field in its final schema-compliant shape
    active: Required[list[external_types.SetPredicate] | None]
    # The `present` field in its final schema-compliant shape
    present: Required[list[external_types.SetPredicate] | None]


class IntermediateActionSelectorDefinition(IntermediateConstructDefinitionBase, total=False):
    """An intermediate action-selector definition.

    The shape captures the structure of action-selector definitions throughout postprocessing.
    """
    # The discriminator for the action-selector construct
    type: Required[Literal[external_types.ConstructDiscriminator.ACTION_SELECTOR]]
    # (Assigned during postprocessing) The `initiator` field in its final schema-compliant shape
    initiator: NotRequired[external_types.RoleName]
    # (Assigned during postprocessing) The `roleForestRoots` field in its final schema-compliant shape
    roleForestRoots: list[external_types.RoleName]
    # (Assigned during postprocessing) The `conditions` field in its final schema-compliant shape
    conditions: external_types.ConstructConditions
    # The `name` field in its final schema-compliant shape
    name: Required[external_types.SelectorName]
    # The `reserved` field in its final schema-compliant shape
    reserved: Required[bool]
    # The `policy` field in its final schema-compliant shape
    policy: Required[external_types.SelectorPolicy]
    # The `candidates` field in its final schema-compliant shape
    candidates: Required[list[external_types.SelectorCandidate]]


class IntermediatePlanSelectorDefinition(IntermediateConstructDefinitionBase, total=False):
    """An intermediate plan-selector definition.

    The shape captures the structure of plan-selector definitions throughout postprocessing.
    """
    # The discriminator for the plan-selector construct
    type: Required[Literal[external_types.ConstructDiscriminator.PLAN_SELECTOR]]
    # (Assigned during postprocessing) The `roleForestRoots` field in its final schema-compliant shape
    roleForestRoots: list[external_types.RoleName]
    # (Assigned during postprocessing) The `conditions` field in its final schema-compliant shape
    conditions: external_types.ConstructConditions
    # The `name` field in its final schema-compliant shape
    name: Required[external_types.SelectorName]
    # The `policy` field in its final schema-compliant shape
    policy: Required[external_types.SelectorPolicy]
    # The `candidates` field in its final schema-compliant shape
    candidates: Required[list[external_types.SelectorCandidate]]


class IntermediateSiftingPatternDefinition(IntermediateConstructDefinitionBase, total=False):
    """An intermediate sifting-pattern definition.

    The shape captures the structure of sifting-pattern definitions throughout postprocessing.
    """
    # The discriminator for the sifting-pattern construct
    type: Required[Literal[external_types.ConstructDiscriminator.SIFTING_PATTERN]]
    # (Assigned during postprocessing) The `roleForestRoots` field in its final schema-compliant shape
    roleForestRoots: list[external_types.RoleName]
    # (Assigned during postprocessing) The `conditions` field in its final schema-compliant shape
    conditions: external_types.ConstructConditions
    # The `name` field in its final schema-compliant shape
    name: Required[external_types.SiftingPatternName]
    # The `actions` field in its final schema-compliant shape
    actions: Required[list[external_types.RoleName]]


class IntermediateTropeDefinition(IntermediateConstructDefinitionBase, total=False):
    """An intermediate trope definition.

    The shape captures the structure of trope definitions throughout postprocessing.
    """
    # The discriminator for the trope construct
    type: Required[Literal[external_types.ConstructDiscriminator.TROPE]]
    # (Assigned during postprocessing) The `roleForestRoots` field in its final schema-compliant shape
    roleForestRoots: list[external_types.RoleName]
    # (Assigned during postprocessing) The `conditions` field in its final schema-compliant shape
    conditions: external_types.ConstructConditions
    # The `name` field in its final schema-compliant shape
    name: Required[external_types.TropeName]


# Union covering all the intermediate construct definitions
IntermediateConstructDefinition = (
    IntermediateActionDefinition
    | IntermediateActionSelectorDefinition
    | IntermediateQueryDefinition
    | IntermediatePlanDefinition
    | IntermediatePlanSelectorDefinition
    | IntermediateSiftingPatternDefinition
    | IntermediateTropeDefinition
)


class IntermediatePlanInstructionJump(TypedDict):
    """An intermediate jump plan instruction, still containing an unresolved target offset."""
    # Discriminator for a jump plan instruction
    type: Literal[external_types.PlanInstructionDiscriminator.JUMP]
    # An offset, relative to the address of this instruction, that will point to the instruction
    # that is the target of this jump. This is resolved once the tape is linearized, which occurs
    # in the plan visitor.
    _target_offset: int


class IntermediatePlanInstructionJumpIfFalse(TypedDict):
    """An intermediate jump-if-false plan instruction, still containing an unresolved target offset."""
    # Discriminator for a jump-if-false plan instruction
    type: Literal[external_types.PlanInstructionDiscriminator.JUMP_IF_FALSE]
    # The `condition` field in its final schema-compliant shape
    _condition: external_types.Expression
    # An offset, relative to the address of this instruction, that will point to the instruction
    # that is the target of this jump. This is resolved once the tape is linearized, which occurs
    # in the plan visitor.
    _target_offset: int


class IntermediatePlanInstructionLoopNext(TypedDict):
    """An intermediate loop-next plan instruction, still containing an unresolved exit-target offset."""
    # Discriminator for a loop-next plan instruction
    type: Literal[external_types.PlanInstructionDiscriminator.LOOP_NEXT]
    # An offset, relative to the address of this instruction, that will point to the instruction
    # to jump to when the loop is exhausted. This is resolved once the tape is linearized, which
    # occurs in the plan visitor.
    _exitTarget_offset: int


class IntermediatePlanConditionalBranch(TypedDict):
    """An intermediate shape representing a branch of a conditional expression in a plan.

    The plan visitor will linearize this into a list of plan instructions.
    """
    # The condition associated with a branch, such that it will be entered if it holds while the
    # condition for no previous branch holds. If this is `None`, it's an alternative branch.
    _condition: external_types.Expression | None
    # A list containing the instructions associated with the branch
    _consequent: list[IntermediatePlanInstruction]


# Union covering all the intermediate plan instructions, most of which are in their final form
IntermediatePlanInstruction = (
    external_types.PlanInstructionAdvance
    | external_types.PlanInstructionFail
    | IntermediatePlanInstructionJump
    | IntermediatePlanInstructionJumpIfFalse
    | external_types.PlanInstructionLoopInit
    | IntermediatePlanInstructionLoopNext
    | external_types.PlanInstructionReactionQueue
    | external_types.PlanInstructionReactionWindowClose
    | external_types.PlanInstructionReactionWindowOpen
    | external_types.PlanInstructionSucceed
    | external_types.PlanInstructionWaitStart
    | external_types.PlanInstructionWaitEnd
)


class IntermediateRoleReference(TypedDict):
    """An intermediate shape containing transient data describing a role reference.

    This transient data does not survive past the visitor step, where it's used for prevalidation.
    """
    # The name of the role being referenced
    _name: external_types.RoleName
    # Whether the entity sigil introduced the reference
    _is_entity_role: bool
    # Whether the reference included the group-role decorator
    _is_group_role: bool


# A sentinel used internally by the compiler
Sentinel: TypeAlias = object



"""Core pipeline orchestrator for the Viv compiler."""

__all__ = ["compile_viv_source_code"]

from pathlib import Path
from importlib.resources import files

import arpeggio
from arpeggio.cleanpeg import ParserPEG

from viv_compiler import (
    bundler,
    config,
    errors,
    external_types,
    includes,
    internal_types,
    postprocessing,
    utils,
    validation,
    visitor
)


def compile_viv_source_code(
    source_file_path: Path,
    default_importance: float,
    default_salience: float,
    default_reaction_priority: float,
    use_memoization: bool,
    verbose_parser: bool = False,
) -> external_types.ContentBundle:
    """Compile the given Viv source file to produce a JSON-serializable compiled content bundle.

    Args:
        source_file_path: The absolute path to the Viv source file to be parsed.
        default_importance: A default importance value to use when one is not specified
            in an action definition.
        default_salience: A default salience value to use when one is not specified
            in an action definition.
        default_reaction_priority: A default reaction priority to use when one is not
            specified in a reaction declaration.
        use_memoization: Whether to use memoization during PEG parsing (faster, but uses more memory).
        verbose_parser: Whether to engage a verbose debugging mode for the PEG parser itself (default: `False`).

    Returns:
        The compiled content bundle.

    Raises:
        VivParseError: The source file could not be parsed.
    """
    # Honor user-supplied config parameters, or the associated default values if none were supplied
    _honor_user_supplied_config_parameters(
        default_importance=default_importance,
        default_salience=default_salience,
        default_reaction_priority=default_reaction_priority,
    )
    # Create a Viv parser
    viv_parser = _create_viv_parser(use_memoization=use_memoization, verbose=verbose_parser)
    # Load the source file to be compiled
    source_file_contents = _load_source_file(source_file_path=source_file_path)
    # Parse the source file to produce a parse tree
    try:
        parse_tree = viv_parser.parse(_input=source_file_contents)
    except arpeggio.NoMatch as parsing_error:
        raise errors.VivParseError(original=parsing_error, file_path=source_file_path) from None
    # Following the visitor pattern in parsing, traverse the parse tree to gradually
    # construct an abstract syntax tree (AST).
    viv_visitor = visitor.Visitor()
    utils.register_source_file(
        source_file_path=source_file_path,
        source_file_contents=source_file_contents,
        position_to_line_and_column=viv_parser.pos_to_linecol
    )
    ast: internal_types.AST = utils.sanitize_ast(
        ast=arpeggio.visit_parse_tree(parse_tree=parse_tree, visitor=viv_visitor)
    )
    # If there are any include declarations (i.e., import statements), honor those now
    combined_ast: internal_types.CombinedAST = includes.integrate_included_files(
        viv_parser=viv_parser,
        viv_visitor=viv_visitor,
        ast=ast,
        entry_point_file_path=source_file_path
    )
    # Conduct postprocessing on the combined AST (modifies it in place)
    postprocessing.postprocess_combined_ast(combined_ast=combined_ast)
    # Package up the compiled content bundle
    compiled_content_bundle = bundler.create_compiled_content_bundle(combined_ast=combined_ast)
    # Conduct final validation and, if it passes, return the validated content bundle. If an issue
    # is detected, an exception will be raised instead.
    validation.validate_content_bundle(content_bundle=compiled_content_bundle)
    return compiled_content_bundle


def _honor_user_supplied_config_parameters(
    default_importance: float,
    default_salience: float,
    default_reaction_priority: float,
) -> None:
    """Updates the global Viv compiler config to honor any user-supplied parameters.

    Args:
        default_importance: A user-provided default importance value to use when one is not specified in
            an action definition. The CLI provides a default value if the user does not supply one.
        default_salience: A user-provided default salience value to use when one is not specified in
            an action definition. The CLI provides a default value if the user does not supply one.
        default_reaction_priority: A user-provided default reaction priority to use when one is not
            specified in a reaction declaration queueing an action or an action selector.

    Returns:
        Nothing. Mutates the config in place.
    """
    config.DEFAULT_IMPORTANCE_SCORE = default_importance
    config.DEFAULT_SALIENCE_SCORE = default_salience
    config.DEFAULT_REACTION_PRIORITY_VALUE = default_reaction_priority


def _create_viv_parser(use_memoization: bool, verbose: bool) -> ParserPEG:
    """Return a PEG parser initialized for the Viv DSL, with user-defined settings.

    Args:
        use_memoization: Whether to use memoization during PEG parsing (faster, but uses more memory).
        verbose: Whether to engage a verbose debugging mode for the PEG parser itself.

    Returns:
        A PEG parser initialized for the Viv DSL.
    """
    # Load the Viv DSL grammar
    viv_grammar = files("viv_compiler.grammar").joinpath("viv.peg").read_text(encoding="utf-8")
    # Prepare and return the parser
    viv_parser = ParserPEG(
        language_def=viv_grammar,
        root_rule_name=config.GRAMMAR_ROOT_SYMBOL,
        comment_rule_name=config.GRAMMAR_COMMENT_SYMBOL,
        reduce_tree=False,
        ws="\t\n\r ",
        memoization=use_memoization,
        debug=verbose
    )
    return viv_parser


def _load_source_file(source_file_path: Path) -> str:
    """Return the contents of the Viv source file at the given path.

    Args:
        source_file_path: Path to the Viv source file to be compiled.

    Returns:
        Contents of the Viv source file at the given path.
    """
    try:
        with open(source_file_path, encoding="utf-8") as source_file:
            return source_file.read()
    except IsADirectoryError:
        raise errors.VivCompileError(f"Expected a file, but got a directory: {source_file_path}") from None
    except FileNotFoundError:
        raise errors.VivCompileError(f"Source file not found: {source_file_path}") from None



"""Mixin class containing visitor methods associated with role definitions."""

__all__ = ["VisitorMixinRoles"]

from math import log
from typing import Any

from arpeggio import NonTerminal, PTNodeVisitor

from viv_compiler import errors, external_types, internal_types, sentinels, utils


class VisitorMixinRoles(PTNodeVisitor):
    """A visitor mixin for Viv roles."""

    @staticmethod
    def visit_role(node: NonTerminal, children: list[Any]) -> external_types.RoleDefinition:
        """Visit a <role> node."""
        # Populate the authored fields
        role_header, role_body = children
        role_header: internal_types.IntermediateRoleReference
        accumulated_fields = {"name": role_header["_name"]}
        role_snippet = VisitorMixinRoles._get_role_error_message_snippet(role_reference=role_header)
        if "renames" in role_body and len(role_body) > 1:
            error_message = f"Role {role_snippet} uses 'renames' alongside other fields ('renames' must be used alone)"
            raise errors.VivCompileError(error_message, source=utils.derive_source_annotations(node=node))
        for field_name, field_value in role_body.items():
            if field_name == "renames":
                parent_role_reference: internal_types.IntermediateRoleReference = field_value
                mismatch = (
                    role_header["_is_entity_role"] != parent_role_reference["_is_entity_role"] or
                    role_header["_is_group_role"] != parent_role_reference["_is_group_role"]
                )
                if mismatch:
                    parent_role_snippet = VisitorMixinRoles._get_role_error_message_snippet(
                        role_reference=parent_role_reference
                    )
                    error_message = f"Sigil mismatch: role {role_snippet} renames {parent_role_snippet}"
                    raise errors.VivCompileError(error_message, source=utils.derive_source_annotations(node=node))
                accumulated_fields["renames"] = parent_role_reference['_name']
            else:
                accumulated_fields[field_name] = field_value
        # Package up the definition
        role_definition = external_types.RoleDefinition(
            name=accumulated_fields['name'],
            entityType=accumulated_fields.get('entityType', external_types.RoleEntityType.CHARACTER),
            min=accumulated_fields.get('min', 1),
            max=accumulated_fields.get('max', 1),
            parent=accumulated_fields.get('parent', None),
            children=accumulated_fields.get('children', []),
            pool=accumulated_fields.get('pool', None),
            chance=accumulated_fields.get('chance', None),
            mean=accumulated_fields.get('mean', None),
            sd=accumulated_fields.get('sd', None),
            participationMode=accumulated_fields.get('participationMode', None),
            anywhere=accumulated_fields.get('anywhere', False),
            precast=accumulated_fields.get('precast', False),
            spawn=accumulated_fields.get('spawn', False),
            spawnFunction=accumulated_fields.get('spawnFunction', None),
            renames=accumulated_fields.get('renames', None),
        )
        # Do some quick validation on transient data
        if role_header["_is_entity_role"]:
            if role_definition['entityType'] == external_types.RoleEntityType.SYMBOL:
                error_message = f"Role {role_snippet} has entity prefix but is marked symbol"
                raise errors.VivCompileError(error_message, source=utils.derive_source_annotations(node=node))
        if not role_header["_is_entity_role"]:
            if role_definition['entityType'] != external_types.RoleEntityType.SYMBOL:
                error_message = f"Role {role_snippet} has symbol prefix but is not marked symbol"
                raise errors.VivCompileError(error_message, source=utils.derive_source_annotations(node=node))
        if role_header["_is_group_role"] and role_definition['max'] < 2:
            error_message = (
                f"Role {role_snippet} has group-role decorator but has a max < 2 "
                "(group roles must be able to cast multiple entities)"
            )
            raise errors.VivCompileError(error_message, source=utils.derive_source_annotations(node=node))
        if not role_header["_is_group_role"] and role_definition['max'] > 1:
            error_message = (
                f"Role {role_snippet} has a max > 1 but is missing group-role decorator '*' "
                f"(rewrite it as {role_snippet}*)"
            )
            raise errors.VivCompileError(error_message, source=utils.derive_source_annotations(node=node))
        return role_definition

    @staticmethod
    def _get_role_error_message_snippet(role_reference: internal_types.IntermediateRoleReference) -> str:
        """Returns a properly decorated role name, for use in error messages."""
        role_snippet = "@" if role_reference["_is_entity_role"] else "&"
        role_snippet += role_reference['_name']
        role_snippet += "*" if role_reference["_is_group_role"] else ""
        return f"'{role_snippet}'"

    @staticmethod
    def visit_role_reference(_, children: list[Any]) -> internal_types.IntermediateRoleReference:
        """Visit a <role_reference> node."""
        binding_type, role_name = children[:2]
        is_entity_role = binding_type is sentinels.ENTITY_BINDING_TYPE
        is_group_role = children[-1] is sentinels.GROUP_ROLE_DECORATOR
        intermediate_role_reference = internal_types.IntermediateRoleReference(
            _name=role_name,
            _is_entity_role=is_entity_role,
            _is_group_role=is_group_role
        )
        return intermediate_role_reference

    @staticmethod
    def visit_binding_type(node: NonTerminal, children: list[Any]) -> object:
        """Visit a <binding_type> node."""
        match children[0]:
            case sentinels.ENTITY_SIGIL:
                return sentinels.ENTITY_BINDING_TYPE
            case sentinels.SYMBOL_SIGIL:
                return sentinels.SYMBOL_BINDING_TYPE
            case _:
                raise errors.VivCompileError(
                    "Encountered unexpected binding type",
                    source=utils.derive_source_annotations(node=node)
                )

    @staticmethod
    def visit_role_body(_, children: list[Any]) -> dict[str, Any]:
        """Visit a <role_body> node."""
        accumulated_fields = {}
        for child in children:
            accumulated_fields.update(child)
        if accumulated_fields.get('mean'):
            # Some initial experimentation has shown that taking the log of max - min produces a solid
            # standard deviation (for a normal distribution centered on the author-supplied mean), but
            # only for smaller values. Once the span eclipses 20 or so, we need a bigger SD to allow
            # for the tails to get better coverage (since otherwise the min and max are effectively
            # ignored). Empirically, it appears that dividing the span by 7 works for bigger spans.
            span = accumulated_fields['max'] - accumulated_fields['min']
            if span != 0:
                sd = max(log(span), span / 7)
            else:
                sd = 0
            accumulated_fields['sd'] = round(sd, 2)
        return accumulated_fields

    @staticmethod
    def visit_role_labels(node: NonTerminal, children: list[str]) -> dict[str, Any]:
        """Visit a <role_labels> node."""
        accumulated_fields = {}
        for role_label in children:
            match role_label:
                case "character":
                    VisitorMixinRoles._set_role_entity_type(
                        node=node,
                        accumulated_fields=accumulated_fields,
                        entity_type=external_types.RoleEntityType.CHARACTER
                    )
                case "item":
                    VisitorMixinRoles._set_role_entity_type(
                        node=node,
                        accumulated_fields=accumulated_fields,
                        entity_type=external_types.RoleEntityType.ITEM
                    )
                case "action":
                    VisitorMixinRoles._set_role_entity_type(
                        node=node,
                        accumulated_fields=accumulated_fields,
                        entity_type=external_types.RoleEntityType.ACTION
                    )
                case "location":
                    VisitorMixinRoles._set_role_entity_type(
                        node=node,
                        accumulated_fields=accumulated_fields,
                        entity_type=external_types.RoleEntityType.LOCATION
                    )
                case "symbol":
                    VisitorMixinRoles._set_role_entity_type(
                        node=node,
                        accumulated_fields=accumulated_fields,
                        entity_type=external_types.RoleEntityType.SYMBOL
                    )
                case "initiator":
                    VisitorMixinRoles._set_role_participation_mode(
                        node=node,
                        accumulated_fields=accumulated_fields,
                        participation_mode=external_types.RoleParticipationMode.INITIATOR
                    )
                case "partner":
                    VisitorMixinRoles._set_role_participation_mode(
                        node=node,
                        accumulated_fields=accumulated_fields,
                        participation_mode=external_types.RoleParticipationMode.PARTNER
                    )
                case "recipient":
                    VisitorMixinRoles._set_role_participation_mode(
                        node=node,
                        accumulated_fields=accumulated_fields,
                        participation_mode=external_types.RoleParticipationMode.RECIPIENT
                    )
                case "bystander":
                    VisitorMixinRoles._set_role_participation_mode(
                        node=node,
                        accumulated_fields=accumulated_fields,
                        participation_mode=external_types.RoleParticipationMode.BYSTANDER
                    )
                case "anywhere":
                    accumulated_fields["anywhere"] = True
                case "precast":
                    accumulated_fields["precast"] = True
                case "spawn":
                    accumulated_fields["spawn"] = True
        return accumulated_fields

    @staticmethod
    def _set_role_entity_type(
        node: NonTerminal,
        accumulated_fields: dict[str, Any],
        entity_type: external_types.RoleEntityType
    ):
        """Sets the given role entity type in the given accumulates role fields, unless one has already been set,
        in which case an informative error is thrown.
        """
        current_entity_type = accumulated_fields.get("entityType", None)
        if not current_entity_type:
            accumulated_fields["entityType"] = entity_type
            return
        error_message = f"Role has multiple entity-type labels: '{current_entity_type}', '{entity_type}'\n"
        raise errors.VivCompileError(error_message, source=utils.derive_source_annotations(node=node))

    @staticmethod
    def _set_role_participation_mode(
        node: NonTerminal,
        accumulated_fields: dict[str, Any],
        participation_mode: external_types.RoleParticipationMode
    ):
        """Sets the given role participation mode in the given accumulated role fields, unless one has already
        been set, in which case an informative error is thrown.
        """
        current_participation_mode = accumulated_fields.get("participationMode", None)
        if not current_participation_mode:
            accumulated_fields["participationMode"] = participation_mode
            return
        error_message = (
            f"Role has multiple participation-mode labels: '{current_participation_mode}' '{participation_mode}'\n"
        )
        raise errors.VivCompileError(error_message, source=utils.derive_source_annotations(node=node))

    @staticmethod
    def visit_role_slots(_, children: list[Any]) -> dict[str, Any]:
        """Visit a <role_slots> node."""
        accumulated_fields = {}
        for child in children:
            accumulated_fields.update(child)
        return accumulated_fields

    @staticmethod
    def visit_role_slots_range(_, children: list[Any]) -> dict[str, int]:
        """Visit a <role_slots_range> node."""
        if len(children) == 1:
            accumulated_fields = {"min": int(children[0]), "max": int(children[0])}
        else:
            accumulated_fields = {"min": int(children[0]), "max": int(children[1])}
        return accumulated_fields

    @staticmethod
    def visit_role_slots_mean(_, children: list[str]) -> dict[str, float]:
        """Visit a <role_slots_mean> node."""
        mean = float(children[0])
        return {"mean": round(mean, 2)}

    @staticmethod
    def visit_role_slots_optional_slot_casting_probability(_, children: list[str]) -> dict[str, float]:
        """Visit a <role_slots_optional_slot_casting_probability> node."""
        chance = float(children[0]) / 100.0
        chance = max(round(chance, 3), 0.001)
        return {"chance": chance}

    @staticmethod
    def visit_role_casting_pool_from(
        _, children: list[external_types.Expression]
    ) -> dict[str, external_types.CastingPool]:
        """Visit a <role_casting_pool_from> node."""
        casting_pool_expression = external_types.CastingPool(body=children[0], uncachable=True)
        return {"pool": casting_pool_expression}

    @staticmethod
    def visit_role_casting_pool_is(
        node: NonTerminal,
        children: list[external_types.Expression]
    ) -> dict[str, external_types.CastingPool]:
        """Visit a <role_casting_pool_is> node."""
        # The `is` is just syntactic sugar for `from` with a singleton array
        casting_pool_expression = external_types.CastingPool(
            body=external_types.ListField(
                type=external_types.ExpressionDiscriminator.LIST,
                value=[children[0]],
                source=utils.derive_source_annotations(node=node)
            ),
            uncachable=True
        )
        return {"pool": casting_pool_expression}

    @staticmethod
    def visit_role_spawn_directive(_, children: list[Any]) -> dict[str, Any]:
        """Visit a <role_spawn_directive> node."""
        accumulated_fields = {
            "spawn": True,
            "spawnFunction": children[0]
        }
        return accumulated_fields

    @staticmethod
    def visit_role_renaming(
        _, children: list[internal_types.IntermediateRoleReference]
    ) -> dict[str, internal_types.IntermediateRoleReference]:
        """Visit a <role_renaming> node."""
        parent_role_reference = children[0]  # Upstream, we will confirm the sigils match and then extract name only
        return {"renames": parent_role_reference}

    @staticmethod
    def visit_entity_sigil(_, __) -> internal_types.Sentinel:
        """Visit an <entity_sigil> node."""
        return sentinels.ENTITY_SIGIL

    @staticmethod
    def visit_symbol_sigil(_, __) -> internal_types.Sentinel:
        """Visit a <symbol_sigil> node."""
        return sentinels.SYMBOL_SIGIL



"""Types associated with the higher-level concerns represented in Viv compiled content bundles."""

from __future__ import annotations

from enum import StrEnum
from typing import TYPE_CHECKING, Literal, TypeAlias, TypedDict

if TYPE_CHECKING:
    from .dsl_types import (
        CustomFunctionCall,
        Enum,
        EnumName,
        Expression,
        FloatField,
        IntField,
        ListField,
        LocalVariable,
        PrecastBindings,
        Reaction,
        SetPredicate,
        StringField,
        TemplateStringField,
        TimeDelta,
        TemporalConstraint
    )


class ContentBundle(TypedDict):
    """A content bundle in the format produced by the Viv compiler."""
    # Metadata for the content bundle, which is currently used for validation purposes
    metadata: ContentBundleMetadata
    # Action definitions, keyed by name
    actions: dict[ActionName, ActionDefinition]
    # Action-selector definitions, keyed by name
    actionSelectors: dict[SelectorName, ActionSelectorDefinition]
    # Plan definitions, keyed by name
    plans: dict[PlanName, PlanDefinition]
    # Query definitions, keyed by name
    queries: dict[QueryName, QueryDefinition]
    # Sifting-pattern definitions, keyed by name
    siftingPatterns: dict[SiftingPatternName, SiftingPatternDefinition]
    # Plan-selector definitions, keyed by name
    planSelectors: dict[SelectorName, PlanSelectorDefinition]
    # Trope definitions, keyed by name
    tropes: dict[TropeName, TropeDefinition]


class ContentBundleMetadata(TypedDict):
    """Metadata on the content bundle.

    This metadata attaches a Viv version number to the content bundle, which guarantees
    compatibility with any Viv runtime with the same version number.

    The metadata here is also used to support validation during initialization of a host application's
    Viv adapter by, e.g., confirming that any referenced enums and adapter functions actually exist.
    """
    # The version number for the Viv content-bundle schema at the time of compiling this
    # content bundle. All Viv runtimes will have a supported version range for content
    # bundles, and will enforce compatibility when a bundle is registered.
    schemaVersion: str
    # The Viv compiler version at the time of compiling this content bundle, stored here
    # to give provenance as a potential aid to debugging.
    compilerVersion: str
    # The Viv DSL grammar version at the time of compiling this content bundle, stored here
    # to give provenance as a potential aid to debugging.
    grammarVersion: str
    # An array containing the names of all enums referenced in the content bundle. This is
    # used for validation during the initialization of a host application's Viv adapter.
    referencedEnums: list[EnumName]
    # An array containing the names of all adapter functions referenced in the
    # content bundle. This is used for validation during the initialization of
    # a host application's Viv adapter.
    referencedFunctionNames: list[str]
    # An array specifying all reactions that are constrained by the time of day. This is used
    # for validation during the initialization of a host application's Viv adapter.
    timeOfDayParameterizedReactions: list[TimeOfDayParameterizedReaction]
    # An array specifying all queries that are parameterized by the time of day. This is used
    # for validation during the initialization of a host application's Viv adapter.
    timeOfDayParameterizedQueries: list[QueryName]
    # A flag indicating whether the content bundle has at least one assignment that modifies entity
    # data. We include this here because such assignments are not allowed in all adapter configurations.
    hasEntityDataAssignments: bool


class TimeOfDayParameterizedReaction(TypedDict):
    """A simple record of a case of reaction that is constrained by the time of day, used for validation purposes."""
    # The type of the construct definition containing a reaction that is constrained by the time of day
    constructType: ConstructDiscriminator
    # The name of the construct definition containing a reaction that is constrained by the time of day
    constructName: ActionName
    # The name of the target construct of the reaction that is constrained by the time of day
    reaction: ActionName | SelectorName | PlanName


class ConstructDiscriminator(StrEnum):
    """Enum specifying the discriminators for all Viv construct types."""
    # Discriminator for action definitions
    ACTION = "action"
    # Discriminator for action-selector definitions
    ACTION_SELECTOR = "actionSelector"
    # Discriminator for plan definitions
    PLAN = "plan"
    # Discriminator for plan-selector definitions
    PLAN_SELECTOR = "planSelector"
    # Discriminator for query definitions
    QUERY = "query"
    # Discriminator for sifting-pattern definitions
    SIFTING_PATTERN = "siftingPattern"
    # Discriminator for trope definitions
    TROPE = "trope"


# Union containing discriminators for the construct types that a reaction may queue
ReactionTargetConstructDiscriminator: TypeAlias = Literal[
    ConstructDiscriminator.ACTION,
    ConstructDiscriminator.ACTION_SELECTOR,
    ConstructDiscriminator.PLAN,
    ConstructDiscriminator.PLAN_SELECTOR
]


class ActionDefinition(TypedDict):
    """A compiled definition for a Viv action."""
    # Discriminator for the action construct type
    type: Literal[ConstructDiscriminator.ACTION]
    # The (unique) name of the action
    name: ActionName
    # Whether this action is reserved, in which case it may only be targeted via selector targeting or queueing
    reserved: bool
    # Mapping from the names of the roles associated with this action to their respective role definitions
    roles: dict[RoleName, RoleDefinition]
    # The name of the initiator role, isolated here for optimization purposes
    initiator: RoleName
    # The names of the roles constituting the roots of the trees composing role-dependency forest for
    # this action definition. The roots are given in the order by which role casting should proceed.
    roleForestRoots: list[RoleName]
    # An expression yielding a numeric importance score the action, for purposes of story sifting
    importance: IntField | FloatField | Enum
    # Tags on the action. These are meant to facilitate search over actions, during story
    # sifting, and their function may be extended in the host application.
    tags: ListField
    # Definition for a simple templated string describing this action in a sentence or so
    gloss: StringField | TemplateStringField | None
    # Definition for a more detailed templated string describing this action in a paragraph or so
    report: StringField | TemplateStringField | None
    # Conditions for the action, grouped by role name (with the special global-conditions key). A condition
    # is an expression that must hold (i.e., evaluate to a truthy value) in order for an action to be performed
    # with a prospective cast.
    conditions: ConstructConditions
    # An ordered set of expressions that prepare a set of temporary variables that may be referenced
    # downstream in the action definition. These temporary variables can be referenced by an author
    # using the `$` sigil, but this is syntactic sugar for `@this.scratch` — e.g., `$&foo` is equivalent
    # to `@this.scratch.foo`, with the second sigil indicating the type of the scratch variable.
    scratch: list[Expression]
    # An ordered set of expressions that, when executed, cause updates to the host application state
    effects: list[WrappedExpression]
    # A set of expressions that each produce a reaction when evaluated. A reaction specifies an action that
    # may be queued up for some time in the future, should an instance of the one at hand be performed.
    reactions: list[WrappedExpression]
    # Specifications for yielding numeric salience values for the action
    saliences: Saliences
    # Specifications for yielding subjective associations for the action
    associations: Associations
    # Embargo directives, which are authorial levers for controlling the frequency
    # with which this action will be performed in the host application.
    embargoes: list[EmbargoDeclaration]


# A unique name for an action
ActionName: TypeAlias = str


class ConstructConditions(TypedDict):
    """Conditions that must hold in order to target a given construct."""
    # Global construct conditions, meaning ones that do not reference any roles
    # and can thus be tested immediately upon targeting a construct.
    globalConditions: list[WrappedExpression]
    # Standard conditions, keyed by the role during whose casting they will be tested
    roleConditions: dict[RoleName, list[WrappedExpression]]


class RoleDefinition(TypedDict):
    """A definition for a role."""
    # A name for this role, unique only within the associated action definition
    name: RoleName
    # The type of entity (or symbol) that must be cast in this role
    entityType: RoleEntityType
    # The minimum number of slots to cast for this role
    min: int
    # The maximum number of slots to cast for this role
    max: int
    # The name of this role's parent, if any, in the dependency tree that is used
    # during role casting. This dependency tree is used to optimize this process.
    parent: RoleName | None
    # The names of this role's children, if any, in the dependency tree that is
    # used during casting. This dependency tree is used to optimize this process.
    children: list[RoleName]
    # If specified, a directive specifying the pool of entities who may be cast into this role at a given
    # point in time, given an initiator and possibly other prospective role bindings. If this is an action
    # role, and if there is an active search domain (from an action search or a sifting), the pool will
    # automatically be filtered to actions in the search domain.
    pool: CastingPool | None
    # If specified, the chance that a qualifying entity will be cast into the role. This field was
    # first implemented to support a pattern of specifying how likely it is that a given nearby
    # character will witness an action, which can be accomplished by defining a `bystander` role
    # with a high `max` and a specified `chance` value. Chance values are always guaranteed to
    # fall between `0.0` and `1.0`.
    chance: float | None
    # A mean on which to anchor a distribution from which will be sampled the number of entities
    # to cast into the role. This distribution must also be parameterized by a `sd` value.
    mean: float | None
    # Standard deviation for a distribution from which will be sampled the number of entities
    # to cast into the role. This distribution must also be parameterized by a `mean` value.
    sd: float | None
    # If applicable, the mode of participation for a character cast in this role. This field
    # is only non-null for action roles that cast characters who are physically present for
    # the action (i.e., ones for which `anywhere` is `false`).
    participationMode: RoleParticipationMode | None
    # Whether an entity cast in this role does not need to be physically present for the action. Note that
    # the entity *can* still be physically present, so authors should take care to write conditions specifying
    # whether an entity must be present, as needed. Currently, this only applies to characters and items,
    # though roles casting all other entity types will have `True` here.
    anywhere: bool
    # Whether this role must be precast and never cast through typical role casting. For an action, a role
    # can be "precast" via a reaction declaration that targets the action. For other constructs, a role
    # can be precast in the expression targeting the construct -- e.g., an action search can precast a role
    # for a query. If an action has a precast role, it must be marked `reserved`, because there is no way
    # to cast a precast role via general action targeting.
    precast: bool
    # Whether the entity cast in this role is to be constructed as a result of the associated
    # action. Spawn roles are always accompanied by an entity recipe.
    spawn: bool
    # For `spawn` roles only, an adapter-function call that will cause the new entity
    # (to be cast in this role) to be constructed before returning its entity ID.
    spawnFunction: CustomFunctionCall | None
    # If this role is an alias for a role in a parent definition, this will
    # store the name of the original role.
    renames: RoleName | None


# A name for an action role (unique only within its action definition)
RoleName: TypeAlias = str


class RoleEntityType(StrEnum):
    """Enum specifying the entity types that may be associated with a role definition."""
    # The role casts an action
    ACTION = "action"
    # The role casts a character
    CHARACTER = "character"
    # The role casts an item
    ITEM = "item"
    # The role casts a location
    LOCATION = "location"
    # The role casts a symbol
    SYMBOL = "symbol"


class RoleParticipationMode(StrEnum):
    """Enum specifying the participation modes for action roles that cast characters.

    To be clear, these are *only* applicable to action roles that cast characters.
    """
    # The role casts a bystander who is an uninvolved witness to the associated action
    BYSTANDER = "bystander"
    # The role casts the single initiator of an action
    INITIATOR = "initiator"
    # The role casts a partner who helps to initiate an action
    PARTNER = "partner"
    # The role casts a recipient of an action
    RECIPIENT = "recipient"


class CastingPool(TypedDict):
    """A directive specifying the pool of entities who may be cast into a role at a given
    point in time, given an initiator and possibly other prospective role bindings.
    """
    # The Viv expression that should evaluate to a casting pool
    body: Expression
    # Whether the casting pool is uncachable. A casting pool is cachable so long as the associated
    # pool declaration *does not* reference a non-initiator role, in which case the role pool would
    # have to be re-computed if the parent role(s) are re-cast (which never happens with an
    # initiator role). When a casting pool is cached, it is not recomputed even as other
    # non-initiator roles are re-cast.
    uncachable: bool


class Saliences(TypedDict):
    """Specifications for determining a numeric salience score for the action that will be held
    by a given character who experiences, observes, or otherwise learns about the action.
    """
    # A specification for a default value to be used as a fallback for any character for which there is no
    # applicable `roles` entry and for which no `custom` expression yielded a value. This will always be
    # structured as a Viv enum, int, or float, where even the enum should resolve to a numeric value.
    default: SalienceScoreExpression
    # A mapping from role names to expressions yielding salience values. For a character who is bound
    # in the given role, the corresponding expression will determine the salience value.
    roles: dict[RoleName, SalienceScoreExpression]
    # For characters for whom no `roles` entry applies, a series of zero or more custom salience-yielding
    # expressions will be evaluated, with the character bound to the local variable specified in the
    # `variable` property. These will be evaluated in turn, with the first numeric evaluated value
    # being assigned as the character's salience. This field is only used if there is no applicable
    # per-role field for the character at hand. If no custom expression evaluates to a numeric value,
    # the default value will be used.
    custom: list[Expression]
    # If there is a non-empty `custom` field, the local variable to which a character will be bound when
    # computing a salience for them. This allows for evaluation of the body expressions, which may refer
    # to this variable in order to do things like conditionalize salience based on the character at hand.
    variable: LocalVariable | None


# An expression that evaluates to a numeric salience score
SalienceScoreExpression: TypeAlias = "Enum | IntField | FloatField"


class Associations(TypedDict):
    """Specifications for determining the subjective associations for the action that will be held
    by a given character who experiences, observes, or otherwise learns about the action.
    """
    # A specification for a default value to be used as a fallback for any character for which there is no
    # applicable `roles` entry and for which no `custom` expression yielded a value. This will always be
    # structured as a Viv list whose elements will be simple Viv string expressions.
    default: ListField
    # A mapping from role names to Viv lists whose elements will be simple Viv string expressions. For a
    # character who is bound in the given role, the corresponding expression will determine their associations.
    roles: dict[RoleName, ListField]
    # For characters for whom no `roles` entry applies, a series of zero or more custom associations-yielding
    # expressions will be evaluated, with the character bound to the local variable specified in the `variable`
    # property. These will be evaluated in turn, with the first evaluated string array being assigned as the
    # character's salience. This field is only used if there is no applicable per-role field for the character
    # at hand. If no custom expression evaluates to a numeric value, the default value will be used.
    custom: list[Expression]
    # If there is a non-empty `custom` field, the local variable to which a character will be bound when
    # computing associations for them. This allows for evaluation of the body expressions, which may refer
    # to this variable in order to do things like conditionalize associations based on the character at hand.
    variable: LocalVariable | None


class EmbargoDeclaration(TypedDict):
    """An embargo declaration constraining the subsequent performance of an associated action."""
    # Names for all the roles constituting the bindings over which this embargo holds. For instance,
    # if two roles R1 and R2 were specified here, and if an action A was performed with bindings
    # R1=[E1] and R2=[E2, E3], then this embargo would hold over all cases of A with any prospective
    # bindings that cast E1 in R1 and *either* E2 and/or E3 in R2. Stated differently, the embargo
    # holds if for all roles specified here, some subset overlaps between the embargo role bindings
    # and the prospective role bindings. Often, an embargo will only specify an initiator.
    roles: list[RoleName] | None
    # Whether the embargo is permanent. If so, `period` will always be null, and exactly one of
    # the fields is guaranteed to be truthy.
    permanent: bool
    # For an embargo that is not permanent, a specification of the time period over which the embargo
    # will hold. If `period` is present, `permanent` will always be false, and exactly one of the
    # fields is guaranteed to be truthy.
    period: TimeDelta | None
    # Whether the embargo holds only over a certain location, that being the location
    # at which an instance of the associated action has just been performed.
    here: bool


class QueryDefinition(TypedDict):
    """A query used to search for actions in a character's memories or in the chronicle."""
    # Discriminator for the query construct type
    type: Literal[ConstructDiscriminator.QUERY]
    # The (unique) name of the query
    name: QueryName
    # Mapping from the names of the roles associated with this query to their respective
    # role definitions. Note that a query may have zero roles, in which case this is empty.
    roles: dict[RoleName, RoleDefinition]
    # The names of the roles constituting the roots of the trees composing role-dependency forest for
    # this query definition. The roots are given in the order by which role casting should proceed.
    roleForestRoots: list[RoleName]
    # Conditions for the query, grouped by role name (with the special global-conditions key). A condition
    # is an expression that must hold (i.e., evaluate to a truthy value) in order for a query to match.
    conditions: ConstructConditions
    # If specified, a component specifying permissible action names for matches to the query
    actionName: list[SetPredicate] | None
    # If specified, a component specifying permissible causal ancestors for matches to the query
    ancestors: list[SetPredicate] | None
    # If specified, a component specifying permissible causal descendants for matches to the query
    descendants: list[SetPredicate] | None
    # If specified, a component specifying a permissible importance range for matches to the query
    importance: QueryNumericRange | None
    # If specified, a component specifying permissible tags for matches to the query
    tags: list[SetPredicate] | None
    # If specified, a component specifying a permissible salience range for matches to the query. If this
    # query is targeted with the chronicle as a search domain, an error will be thrown, since the query
    # applies to character memories.
    salience: QueryNumericRange | None
    # If specified, a component specifying permissible associations for matches to the query. If this
    # query is targeted with the chronicle as a search domain, an error will be thrown, since the query
    # applies to character memories.
    associations: list[SetPredicate] | None
    # If specified, a component specifying permissible locations for matches to the query
    location: list[SetPredicate] | None
    # If specified, a component specifying permissible performance times for matches to the query
    time: list[TemporalConstraint] | None
    # If specified, a component specifying permissible initiator-role bindings for matches to the query
    initiator: list[SetPredicate] | None
    # If specified, a component specifying permissible partner-role bindings for matches to the query
    partners: list[SetPredicate] | None
    # If specified, a component specifying permissible recipient-role bindings for matches to the query
    recipients: list[SetPredicate] | None
    # If specified, a component specifying permissible bystander-role bindings for matches to the query
    bystanders: list[SetPredicate] | None
    # If specified, a component specifying permissible active-role bindings for matches to the query
    active: list[SetPredicate] | None
    # If specified, a component specifying permissible present-role bindings for matches to the query
    present: list[SetPredicate] | None


# A unique name for a query
QueryName: TypeAlias = str


class QueryNumericRange(TypedDict):
    """A permissible numeric range used in a query."""
    # The lower bound of the range, if there is one, else `None`
    lower: QueryNumericRangeBound | None
    # The upper bound of the range, if there is one, else `None`
    upper: QueryNumericRangeBound | None


class QueryNumericRangeBound(TypedDict):
    """A bound on a permissible numeric range used in a query."""
    # The number at this bound of a range
    value: IntField | FloatField | Enum
    # Whether this bound is inclusive
    inclusive: bool


class PlanDefinition(TypedDict):
    """The definition for a plan, which an author defines to orchestrate complex action sequences
    that may play out over extended periods of story time.

    A plan is composed primarily by an ordered set of phases, each of which centers on a *tape* of plan
    instructions. The plan executor executes a phase by stepping a program counter through the phase tape,
    which specifies control flow along with instructions for queueing the material that makes up the plan.
    """
    # Discriminator for the plan construct type
    type: Literal[ConstructDiscriminator.PLAN]
    # The (unique) name of the plan
    name: PlanName
    # Mapping from the names of the roles associated with this plan to their respective role
    # definitions. Note that the roles appear in the order in which the author defined them.
    roles: dict[RoleName, RoleDefinition]
    # The names of the roles constituting the roots of the trees composing role-dependency forest for this
    # plan definition. The roots are given in the order by which role casting will proceed.
    roleForestRoots: list[RoleName]
    # Conditions for the plan, grouped by role name (with the special global-conditions key). A condition is
    # an expression that must hold (i.e., evaluate to a truthy value) in order for the plan to be launched.
    conditions: ConstructConditions
    # A mapping from phase name to phase definition, for all the phases structuring the plan. Note that
    # the individual phases each point to the next phase, so we don't need to maintain order here.
    phases: dict[PlanPhaseName, PlanPhase]
    # The name of the initial phase in the plan. Plan execution will always begin in this phase.
    initialPhase: PlanPhaseName


# A unique name for a plan
PlanName: TypeAlias = str


class PlanPhase(TypedDict):
    """A phase in a plan, which is structured as a tape of instructions for which execution can
    be arbitrary paused according to the author-defined control flow.

    For instance, an author can specify that the plan will not resume until some period of story time
    has elapsed, or until one of the reactions in a group has been performed, and so forth.
    """
    # The name for the phase, guaranteed to be unique (only) within the enclosing plan
    name: PlanPhaseName
    # The name of the next phase in the plan, if any, else `None` if it's the last phase in the
    # plan. Completion of a final phase causes the plan to resolve with a final success status.
    next: PlanPhaseName | None
    # A list containing the compiled instruction tape for this phase, such that each array index serves
    # as an address. If execution reaches the end of the tape, execution proceeds to the next phase in
    # the plan, if there is one; otherwise, the plan succeeds.
    tape: list[PlanInstruction]


# A name for a plan phase, guaranteed to be unique only within the enclosing plan
PlanPhaseName: TypeAlias = str


class PlanInstructionAdvance(TypedDict):
    """A plan instruction that advances to the next plan phase, skipping any remaining
    instructions in the current phase.
    """
    # Discriminator for an advance plan instruction
    type: Literal[PlanInstructionDiscriminator.ADVANCE]


class PlanInstructionFail(TypedDict):
    """A plan instruction that resolves the plan with a final failure status."""
    # Discriminator for a fail plan instruction
    type: Literal[PlanInstructionDiscriminator.FAIL]


class PlanInstructionJump(TypedDict):
    """A plan instruction that causes an unconditional jump to a different position
    in the phase's instruction tape.
    """
    # Discriminator for a jump plan instruction
    type: Literal[PlanInstructionDiscriminator.JUMP]
    # The instruction address to which execution will jump unconditionally
    target: PlanInstructionAddress


class PlanInstructionJumpIfFalse(TypedDict):
    """A plan instruction that specifies a conditional jump to a different position in the phase's
    instruction tape.

    If the condition does not hold, the plan executor will proceed to the next instruction
    (i.e., increment program counter by one). This enables conditionals in plan bodies.
    """
    # Discriminator for a jump-if-false plan instruction
    type: Literal[PlanInstructionDiscriminator.JUMP_IF_FALSE]
    # The condition to evaluate. If it does *not* hold, execution will proceed to the instruction address specified
    # in the `target` field, else it will proceed to the next instruction (by incrementing the counter by one).
    condition: Expression
    # The instruction address to which execution will jump if `condition` does not hold
    target: PlanInstructionAddress


class PlanInstructionLoopInit(TypedDict):
    """A plan instruction that initializes a loop frame."""
    # Discriminator for a loop-init plan instruction
    type: Literal[PlanInstructionDiscriminator.LOOP_INIT]
    # The expression yielding the iterable to loop over. This will be evaluated once, at the time
    # the loop is initialized. For example, a loop over `@foo.friends` with a loop-body instruction
    # causing a one-year wait after each iteration would still be operating over the friends that
    # `@foo` had at the time the loop was entered (years ago, in diegetic terms).
    iterable: Expression
    # The local variable to bind on each iteration, which the plan executor stores in the loop
    # frame that will be created when executing this instruction.
    variable: LocalVariable


class PlanInstructionLoopNext(TypedDict):
    """A plan instruction that advances an active loop frame by one iteration, or else exits
    the loop if the iterable has been exhausted.

    If there is a next element in the iterable, the plan executor will bind it to the loop variable and
    fall through to the next instruction (i.e., increment by the program counter by one). If the iterable
    has been exhausted, the executor will jump to the instruction address specified by `exitTarget`.
    """
    # Discriminator for a loop-next plan instruction
    type: Literal[PlanInstructionDiscriminator.LOOP_NEXT]
    # The instruction address to jump to when the loop is exhausted. If the iterable is empty
    # from the start, the loop will immediately exit by virtue of this instruction.
    exitTarget: PlanInstructionAddress


class PlanInstructionReactionQueue(TypedDict):
    """A plan instruction that causes queueing of an action, plan, or selector according to a specified reaction."""
    # Discriminator for a reaction-queue plan instruction
    type: Literal[PlanInstructionDiscriminator.REACTION_QUEUE]
    # The reaction declaration
    reaction: Reaction


class PlanInstructionReactionWindowOpen(TypedDict):
    """A plan instruction that opens a new reaction window whose resolution will be governed by
    an operator specified in the corresponding window-close instruction."""
    # Discriminator for a plan instruction that opens a reaction window
    type: Literal[PlanInstructionDiscriminator.REACTION_WINDOW_OPEN]


class PlanInstructionReactionWindowClose(TypedDict):
    """A plan instruction that closes the active reaction window, pending resolution that is
    governed by a specified operator."""
    # Discriminator for a plan instruction that closes the active reaction window
    type: Literal[PlanInstructionDiscriminator.REACTION_WINDOW_CLOSE]
    # The logical operator that will govern resolution of the reaction window
    operator: PlanPhaseReactionWindowOperator


class PlanInstructionSucceed(TypedDict):
    """A plan instruction that resolves the plan with a final success status."""
    # Discriminator for a succeed plan instruction
    type: Literal[PlanInstructionDiscriminator.SUCCEED]


class PlanInstructionWaitEnd(TypedDict):
    """A plan instruction that resumes plan execution following a pause imposed by a wait-start instruction."""
    # Discriminator for a wait-end plan instruction
    type: Literal[PlanInstructionDiscriminator.WAIT_END]
    # If specified, a set of expressions that must hold to resume plan execution prior to the timeout elapsing
    resumeConditions: list[Expression] | None


class PlanInstructionWaitStart(TypedDict):
    """A plan instruction that commences a pause on plan execution that will persist
    until a timeout elapses or a set of conditions hold.
    """
    # Discriminator for a wait plan instruction
    type: Literal[PlanInstructionDiscriminator.WAIT_START]
    # A timeout, expressed as the maximum period of story time to wait counting from the time
    # at which this instruction was executed.
    timeout: TimeDelta


# An address for an instruction in a plan phase, which is really just as an index into the
# phase's `tape` array property. As plan execution proceeds, the executor maintains the
# address for the current instruction as a program counter.
PlanInstructionAddress = int


class PlanInstructionDiscriminator(StrEnum):
    """Discriminator values for compiled plan instructions."""
    # Advance to the next phase in the plan
    ADVANCE = "advance"
    # Resolve the plan with a final failure status
    FAIL = "fail"
    # Jump unconditionally to a new plan instruction
    JUMP = "jump"
    # Jump to a new plan instruction if a condition does not hold
    JUMP_IF_FALSE = "jumpIfFalse"
    # Initialize a `foreach`-style loop frame
    LOOP_INIT = "loopInit"
    # Advance an active loop frame to the next iteration, or exit the loop if its iterations have been exhausted
    LOOP_NEXT = "loopNext"
    # Queue a single reaction
    REACTION_QUEUE = "reactionQueue"
    # Open a new reaction window
    REACTION_WINDOW_OPEN = "reactionWindowOpen"
    # Close the active reaction window, and start trying to resolve the window according to its operator
    REACTION_WINDOW_CLOSE = "reactionWindowClose"
    # Resolve the plan with a final success status
    SUCCEED = "succeed"
    # Resume execution of the plan following a pause imposed by a wait-start instruction
    WAIT_END = "waitEnd"
    # Commence a pause on execution of the plan that will persist until a timeout occurs (diegetic duration),
    # or until an optional set of author-supplied conditions hold.
    WAIT_START = "waitStart"


class PlanPhaseReactionWindowOperator(StrEnum):
    """Enum containing the valid operators for reaction windows.

    When a reaction window is active during execution of a plan phase, all reactions that are queued during
    the window are tracked. Once the end of the window is reached, execution focuses on resolving the window
    according to its operator. Depending on the operator and the reaction outcomes, this may require all the
    queued reactions to be resolved.

    Note: The DSL allows for authors to use the `untracked` operator, but this is equivalent to declaring
    a sequence of bare reactions, and in fact this is how an `untracked` window is compiled. As such, the
    operator never makes it into the compiled content bundle, hence its not appearing here.
    """
    # All the reactions must be performed/launched in order for plan execution to proceed. Note that
    # plan execution will always proceed if no reactions are queued during the reaction window.
    ALL = "all"
    # At least one of the reactions must be performed/launched in order for plan execution to proceed. Note
    # that the plan will always fail if no reactions are queued during the reaction window.
    ANY = "any"


# A Viv plan instruction
PlanInstruction = (
        PlanInstructionAdvance
        | PlanInstructionFail
        | PlanInstructionJump
        | PlanInstructionJumpIfFalse
        | PlanInstructionLoopInit
        | PlanInstructionLoopNext
        | PlanInstructionReactionQueue
        | PlanInstructionReactionWindowClose
        | PlanInstructionReactionWindowOpen
        | PlanInstructionSucceed
        | PlanInstructionWaitStart
        | PlanInstructionWaitEnd
)


class SelectorDefinitionBase(TypedDict):
    """Base class for action-selector definitions and plan-selector definitions."""
    # The (unique) name of the selector
    name: SelectorName
    # Mapping from the names of the roles associated with this selector to their respective role
    # definitions. Note that the roles appear in the order in which the author defined them.
    roles: dict[RoleName, RoleDefinition]
    # The names of the roles constituting the roots of the trees composing role-dependency forest for this
    # selector definition. The roots are given in the order by which role casting will proceed.
    roleForestRoots: list[RoleName]
    # Conditions for the selector, grouped by role name (with the special global-conditions key). A condition is
    # an expression that must hold (i.e., evaluate to a truthy value) in order for the selector to be targeted.
    conditions: ConstructConditions
    # The sort policy that will be used to determine the order in which candidates will be targeted
    policy: SelectorPolicy
    # The candidates (actions and/or other action selectors) that may be targeted via this selector
    candidates: list[SelectorCandidate]


# A unique name for a selector
SelectorName: TypeAlias = str


class SelectorPolicy(StrEnum):
    """Enum containing the valid selector policies."""
    # Target the candidates in the author-specified order
    ORDERED = "ordered"
    # Target the candidates in random order
    RANDOMIZED = "randomized"
    # Target the candidates in weighted random order
    WEIGHTED = "weighted"


class SelectorCandidate(TypedDict):
    """A candidate that may be targeted via a selector."""
    # The name of the candidate, i.e., the name of the construct to target
    name: ActionName | PlanName | SelectorName
    # Whether the candidate is a selector. The compiler ensures that action selectors can only
    # target other action selectors, and that plan selectors can only target other plan selectors.
    isSelector: bool
    # Precast bindings for the candidate, as asserted in the selector definition
    bindings: PrecastBindings
    # If applicable, an expression that will evaluate to the weight for this candidate,
    # which will be used as part of a weighted random sort procedure.
    weight: Expression | None


class ActionSelectorDefinition(SelectorDefinitionBase):
    """The definition for an action selector, which groups candidate actions (and potentially other action
    selectors) under a targeting policy and succeeds upon successful targeting of one of the candidates.
    """
    # Discriminator for the action-selector construct type
    type: Literal[ConstructDiscriminator.ACTION_SELECTOR]
    # Whether this selector is reserved, in which case it may only be targeted via selector targeting or queueing
    reserved: bool
    # The name of the initiator role, isolated here for optimization purposes
    initiator: RoleName


class PlanSelectorDefinition(SelectorDefinitionBase):
    """The definition for a plan selector, which groups candidate plans (and potentially other plan
    selectors) under a targeting policy and succeeds upon successful launching of one of the candidates.
    """
    # Discriminator for the plan-selector construct type
    type: Literal[ConstructDiscriminator.PLAN_SELECTOR]


class SiftingPatternDefinition(TypedDict):
    """The definition for a sifting pattern, which is used to retrieve a sequence of
    actions that together may be construed as constituting the events in a story.
    """
    # Discriminator for the sifting-pattern construct type
    type: Literal[ConstructDiscriminator.SIFTING_PATTERN]
    # The (unique) name of the sifting pattern
    name: SiftingPatternName
    # Mapping from the names of the roles associated with this sifting pattern to their respective
    # role definitions. This will include the role definitions defined in the 'actions' section of a
    # sifting-pattern definition, whose role names will also be stored in the `actions` property here.
    roles: dict[RoleName, RoleDefinition]
    # The names of the roles constituting the roots of the trees composing role-dependency forest for this
    # sifting-pattern definition. The roots are given in the order by which role casting should proceed.
    roleForestRoots: list[RoleName]
    # Conditions for the sifting pattern, grouped by role name (with the special global-conditions key). A condition
    # is an expression that must hold (i.e., evaluate to a truthy value) in order for the sifting pattern to match.
    conditions: ConstructConditions

    # An array containing the roles names for all the actions to expose in a match for this pattern. These will
    # correspond to a subset of the roles defined in the `roles` property.
    actions: list[RoleName]


# A unique name for a sifting pattern
SiftingPatternName: TypeAlias = str


class TropeDefinition(TypedDict):
    """A definition for a Viv trope (reusable bundle of conditions)."""
    # Discriminator for the trope construct type
    type: Literal[ConstructDiscriminator.TROPE]
    # The (unique) name of the trope
    name: TropeName
    # Mapping from the names of the roles associated with this trope to their respective role definitions
    roles: dict[RoleName, RoleDefinition]
    # The names of the roles constituting the roots of the trees composing role-dependency forest for
    # this trope definition. The roots are given in the order by which role casting should proceed.
    roleForestRoots: list[RoleName]
    # Conditions for the trope, grouped by role name (with the special global-conditions key). A condition is an
    # expression that must hold (i.e., evaluate to a truthy value) in order for the trope to fit a given cast.
    conditions: ConstructConditions


# A unique name for a trope
TropeName: TypeAlias = str


# A Viv construct definition
ConstructDefinition = (
    ActionDefinition
    | ActionSelectorDefinition
    | QueryDefinition
    | PlanDefinition
    | PlanSelectorDefinition
    | SiftingPatternDefinition
    | TropeDefinition
)


class WrappedExpression(TypedDict):
    """A Viv expression wrapped with an array containing the names of all roles that it references.

    These reference lists are used for various optimizations.
    """
    # The actual Viv expression that is being wrapped
    body: Expression
    # Names of the roles referenced in the Viv expression
    references: list[RoleName]



"""Mixin class containing visitor methods associated with reaction declarations."""

__all__ = ["VisitorMixinReactions"]

from typing import Any

from arpeggio import NonTerminal, PTNodeVisitor

from viv_compiler import config, errors, external_types, internal_types, sentinels, utils


class VisitorMixinReactions(PTNodeVisitor):
    """A visitor mixin for Viv reactions."""

    @staticmethod
    def visit_reaction(node: NonTerminal, children: list[Any]) -> external_types.Reaction:
        """Visit a <reaction> node."""
        # Populate the fields that were authored
        accumulated_fields = {}
        for child in children:
            accumulated_fields.update(child)
        if "bindings" not in accumulated_fields:
            raise errors.VivCompileError(
                "Reaction has no bindings (use 'with none;' to precast zero roles)",
                source=utils.derive_source_annotations(node=node)
            )
        # If the reaction targets an action or an action selector, prepare a default priority
        priority_default = None
        target_type = accumulated_fields['targetType']
        if target_type in (
            external_types.ConstructDiscriminator.ACTION,
            external_types.ConstructDiscriminator.ACTION_SELECTOR
        ):
            priority_default = external_types.FloatField(
                type=external_types.ExpressionDiscriminator.FLOAT,
                value=config.DEFAULT_REACTION_PRIORITY_VALUE,
                source=None
            )
        # Package up the definition
        reaction_value = external_types.Reaction(
            type=external_types.ExpressionDiscriminator.REACTION,
            value=external_types.ReactionValue(
                targetName=accumulated_fields['targetName'],
                targetType=target_type,
                bindings=accumulated_fields['bindings'],
                urgent=accumulated_fields.get('urgent', None),
                priority=accumulated_fields.get('priority', priority_default),
                location=accumulated_fields.get('location', None),
                time=accumulated_fields.get('time', None),
                repeatLogic=accumulated_fields.get('repeatLogic', None),
                abandonmentConditions=accumulated_fields.get('abandonmentConditions', None),
            ),
            source=utils.derive_source_annotations(node=node)
        )
        return reaction_value

    @staticmethod
    def visit_reaction_header(_, children: list[Any]) -> dict[str, Any]:
        """Visit a <reaction_header> node."""
        accumulated_fields = {}
        for child in children:
            accumulated_fields.update(child)
        return accumulated_fields

    @staticmethod
    def visit_reaction_target(node: NonTerminal, children: list[internal_types.Sentinel | str]) -> dict[str, Any]:
        """Visit a <reaction_target> node."""
        target_type, target_name = children
        if target_type is sentinels.REACTION_TARGET_TYPE_ACTION:
            target_type = external_types.ConstructDiscriminator.ACTION
        elif target_type is sentinels.REACTION_TARGET_TYPE_ACTION_SELECTOR:
            target_type = external_types.ConstructDiscriminator.ACTION_SELECTOR
        elif target_type is sentinels.REACTION_TARGET_TYPE_PLAN:
            target_type = external_types.ConstructDiscriminator.PLAN
        elif target_type is sentinels.REACTION_TARGET_TYPE_PLAN_SELECTOR:
            target_type = external_types.ConstructDiscriminator.PLAN_SELECTOR
        else:
            raise errors.VivCompileError(
                "Unexpected construct type for reaction target",
                source=utils.derive_source_annotations(node=node)
            )
        accumulated_fields = {
            "targetType": target_type,
            "targetName": target_name
        }
        return accumulated_fields

    @staticmethod
    def visit_reaction_target_type(node: NonTerminal, _) -> internal_types.Sentinel:
        """Visit a <reaction_target_type> node."""
        match str(node):
            case "action":
                return sentinels.REACTION_TARGET_TYPE_ACTION
            case "action-selector":
                return sentinels.REACTION_TARGET_TYPE_ACTION_SELECTOR
            case "plan":
                return sentinels.REACTION_TARGET_TYPE_PLAN
            case "plan-selector":
                return sentinels.REACTION_TARGET_TYPE_PLAN_SELECTOR
            case _:
                raise errors.VivCompileError(
                    "Unexpected construct type for reaction target",
                    source=utils.derive_source_annotations(node=node)
                )

    @staticmethod
    def visit_reaction_body(_, children: list[Any]) -> dict[str, Any]:
        """Visit a <reaction_body> node."""
        accumulated_fields = {}
        for child in children:
            accumulated_fields.update(child)
        return accumulated_fields

    @staticmethod
    def visit_reaction_urgency(_, children: list[external_types.Expression]) -> dict[str, external_types.Expression]:
        """Visit a <reaction_urgency> node."""
        return {"urgent": children[0]}

    @staticmethod
    def visit_reaction_priority(_, children: list[external_types.Expression]) -> dict[str, external_types.Expression]:
        """Visit a <reaction_priority> node."""
        return {"priority": children[0]}

    @staticmethod
    def visit_reaction_location(
        _, children: list[external_types.SetPredicate]
    ) -> dict[str, list[external_types.SetPredicate]]:
        """Visit a <reaction_location> node."""
        return {"location": children}

    @staticmethod
    def visit_reaction_time(
        _, children: list[external_types.TemporalConstraint]
    ) -> dict[str, list[external_types.TemporalConstraint]]:
        """Visit a <reaction_time> node."""
        return {"time": children}

    @staticmethod
    def visit_reaction_abandonment_conditions(
        _, children: list[list[external_types.Expression]]
    ) -> dict[str, list[external_types.Expression]]:
        """Visit a <reaction_abandonment_conditions> node."""
        return {"abandonmentConditions": children[0]}

    @staticmethod
    def visit_reaction_repeat_logic(_, children: list[Any]) -> dict[str, external_types.ReactionRepeatLogic]:
        """Visit a <reaction_repeat_logic> node."""
        accumulated_fields = {}
        for child in children:
            accumulated_fields.update(child)
        reaction_repeat_logic = external_types.ReactionRepeatLogic(
            conditions=accumulated_fields["conditions"],
            maxInstances=accumulated_fields["maxInstances"]
        )
        return {"repeatLogic": reaction_repeat_logic}

    @staticmethod
    def visit_reaction_repeat_logic_if(
        _, children: list[list[external_types.Expression]]
    ) -> dict[str, list[external_types.Expression]]:
        """Visit a <reaction_repeat_logic_if> node."""
        return {"conditions": children[0]}

    @staticmethod
    def visit_reaction_repeat_logic_max(_, children: list[Any]) -> dict[str, int]:
        """Visit a <reaction_repeat_logic_max> node."""
        return {"maxInstances": int(children[0])}



"""Handles validation of the compiled content bundle."""

__all__ = ["validate_content_bundle"]

from viv_compiler import external_types
from ._actions import validate_action_definitions
from ._plans import validate_plan_definitions
from ._queries import validate_query_definitions
from ._selectors import validate_action_selector_definitions, validate_plan_selector_definitions
from ._sifting_patterns import validate_sifting_pattern_definitions
from ._tropes import validate_trope_definitions


def validate_content_bundle(content_bundle: external_types.ContentBundle) -> None:
    """Validate the given compiled content bundle.

    Args:
        content_bundle: The compiled content bundle to validate.

    Returns:
        None. Only returns if no issue was detected downstream.
    """
    # Validate action definitions
    validate_action_definitions(content_bundle=content_bundle)
    # Validate action-selector definitions
    validate_action_selector_definitions(content_bundle=content_bundle)
    # Validate plan definitions
    validate_plan_definitions(content_bundle=content_bundle)
    # Validate plan-selector definitions
    validate_plan_selector_definitions(content_bundle=content_bundle)
    # Validate query definitions
    validate_query_definitions(content_bundle=content_bundle)
    # Validate sifting-pattern definitions
    validate_sifting_pattern_definitions(content_bundle=content_bundle)
    # Validate trope definitions
    validate_trope_definitions(content_bundle=content_bundle)



"""Command‑line interface (CLI) for the Viv compiler."""

__all__ = ["main"]

import os
import sys
import json
import argparse
import traceback
from contextlib import contextmanager
from importlib import resources
from pathlib import Path
from typing import Any

from viv_compiler import (
    __version__,
    __schema_version__,
    __grammar_version__,
    api,
    config,
    errors,
    external_types
)


def main() -> None:
    """Command-line interface (CLI) for the Viv DSL compiler."""
    # Build the parser for command-line arguments
    command_line_parser = _build_parser()
    # Parse the command-line arguments
    args = command_line_parser.parse_args()
    # If the user has requested compiler version numbers, print them and exit
    if args.version:
        _print_versions()
        sys.exit(0)
    with _handle_compiler_errors(show_traceback=args.traceback):
        # If test mode is not engaged and no source file was provided, error and exit
        if not args.test and not args.input:
            command_line_parser.error("Unless the 'test' flag is engaged, a Viv source file must be provided")
        # If test mode is engaged, invoke the compiler on a test file and exit
        if args.test:
            _run_smoke_test()
            sys.exit(0)
        # Otherwise, it's showtime, so let's invoke the compiler
        compiled_content_bundle = _invoke_compiler(args=args)
        # If we get to here, compilation succeeded
        if not args.quiet:
            print(_bold(_green("\n* Compilation succeeded")), file=sys.stderr)
        if args.output:
            path_to_output_file = Path(args.output).expanduser().resolve()
            if not path_to_output_file.parent.exists():
                raise errors.VivCompileError(f"Output-file directory does not exist: {path_to_output_file.parent}")
        else:
            path_to_output_file = None
        _emit_results(
            content_bundle=compiled_content_bundle,
            args=args,
            path_to_output_file=path_to_output_file,
            quiet=args.quiet
        )


def _build_parser() -> argparse.ArgumentParser:
    """Build the parser for our command-line arguments.

    Returns:
        A prepared parser for our command-line arguments.
    """
    parser = argparse.ArgumentParser(
        description="Compile a Viv source file (.viv) to produce a content bundle ready for use in a Viv runtime",
        formatter_class=argparse.ArgumentDefaultsHelpFormatter
    )
    parser.add_argument(
        '-V',
        '--version',
        action='store_true',
        help='print compiler, schema, and grammar versions and exit'
    )
    parser.add_argument(
        '-i',
        '--input',
        metavar='PATH',
        type=str,
        help='relative or absolute path to the Viv source file (.viv) to be compiled'
    )
    parser.add_argument(
        '-o',
        '--output',
        metavar='PATH',
        type=str,
        default=None,
        help='relative or absolute path to which to write an output file (.json)'
    )
    parser.add_argument(
        '-p',
        '--print',
        metavar='FILTER',
        type=str,
        nargs='?',
        const='',  # Bare `-p` with no argument
        default=None,
        help=(
            "print compiled output, optionally with filter (examples: 'plans', "
            "'plans.plot-revenge', 'plans.plot-revenge.roles.plotter')"
        )
    )
    parser.add_argument(
        '-l',
        '--list',
        action='store_true',
        default=False,
        help='after compilation, list all compiled construct names in console'
    )
    parser.add_argument(
        '--default-importance',
        metavar='FLOAT',
        type=float,
        default=config.DEFAULT_IMPORTANCE_SCORE,
        help='default importance value to use when one is not specified'
    )
    parser.add_argument(
        '--default-salience',
        metavar='FLOAT',
        type=float,
        default=config.DEFAULT_SALIENCE_SCORE,
        help='default salience value to use when one is not specified'
    )
    parser.add_argument(
        '--default-reaction-priority',
        metavar='FLOAT',
        type=float,
        default=config.DEFAULT_REACTION_PRIORITY_VALUE,
        help='default reaction priority to use when not specified'
    )
    parser.add_argument(
        '--memoization',
        action=argparse.BooleanOptionalAction,
        default=True,
        help='enable/disable memoization in the underlying PEG parser',
    )
    parser.add_argument(
        '--verbose-parser',
        action='store_true',
        default=False,
        help=(
            'engage a verbose debug mode in the underlying PEG parser '
            '(side effect: writes debug files to working directory)'
        )
    )
    parser.add_argument(
        '--traceback',
        action='store_true',
        default=False,
        help='upon a compiler error, log traceback in addition to the error message'
    )
    parser.add_argument(
        '--test',
        action='store_true',
        default=False,
        help='run a simple smoke test to confirm the compiler is installed correctly'
    )
    parser.add_argument(
        '-q',
        '--quiet',
        action='store_true',
        default=False,
        help='suppress status output on success (errors will still be printed)'
    )
    return parser


def _print_versions() -> None:
    """Print version numbers for the compiler package, content-bundle schema, and DSL grammar."""
    versions = (
        "\n* Installed versions:"
        # Version number for the `viv_compiler` package itself
        f"\n\n  - Compiler Package: {__version__}"
        # Version number for the content-bundle schema
        f"\n\n  - Schema:  {__schema_version__}"
        # Version number for the Viv DSL grammar (editor plugins assume compatibility)
        f"\n\n  - Grammar: {__grammar_version__}\n"
    )
    print(_cyan(versions))


def _run_smoke_test() -> None:
    """Runs a smoke test to confirm that the Viv compiler installation appears to be functioning.

    Side Effects:
        The results are printed out to `stderr`.
    """
    print(_cyan("\n* Compiling sample file..."), file=sys.stderr)
    with resources.as_file(resources.files("viv_compiler._samples") / "smoke-test.viv") as sample_path:
        api.compile_from_path(source_file_path=sample_path)
        print(_bold(_green("\n* Smoke test passed")), file=sys.stderr)
        print(_green("\nViv compiler installation is operational.\n"), file=sys.stderr)


def _invoke_compiler(args: argparse.Namespace) -> external_types.ContentBundle:
    """Invokes the compiler on the user's specified source file, with their specified configuration settings.

    Args:
        args: Parsed command-line arguments.

    Returns:
        The compiled content bundle, if compilation succeeds.
    """
    source_file_path = Path(args.input).expanduser().resolve()
    if not args.quiet:
        print(_cyan(f"\n* Compiling source file: {source_file_path.name}"), file=sys.stderr)
    compiled_content_bundle = api.compile_from_path(
        source_file_path=source_file_path,
        default_importance=args.default_importance,
        default_salience=args.default_salience,
        default_reaction_priority=args.default_reaction_priority,
        use_memoization=args.memoization,
        verbose_parser=args.verbose_parser,
    )
    return compiled_content_bundle


@contextmanager
def _handle_compiler_errors(show_traceback: bool):
    """Context manager that catches and formats compiler errors for CLI output."""
    # User interrupts
    try:
        yield
    except KeyboardInterrupt:
        sys.exit(130)
    except BrokenPipeError:
        # noinspection PyBroadException
        try:
            sys.stderr.close()
        except Exception:
            pass
        sys.exit(1)
    # Parsing errors
    except errors.VivParseError as parse_error:
        print(_bold(_red(f"\n* Compilation failed:")), file=sys.stderr)
        print(_red(f"\n{parse_error}\n"), file=sys.stderr)
        sys.exit(1)
    # Compilation errors
    except errors.VivCompileError as compile_error:
        if show_traceback:
            traceback_str = "".join(
                traceback.format_exception(type(compile_error), compile_error, compile_error.__traceback__)
            )
            print(_yellow(f"\n{traceback_str}"), file=sys.stderr)
        print(_bold(_red(f"\n* Compilation failed:")), file=sys.stderr)
        print(_red(f"\n{compile_error}\n"), file=sys.stderr)
        if show_traceback:
            print(_red(f"Scroll up for the full traceback.\n"), file=sys.stderr)
        sys.exit(1)
    # Some kind of unexpected error
    except Exception as internal_error:
        traceback_str = "".join(
            traceback.format_exception(type(internal_error), internal_error, internal_error.__traceback__)
        )
        print(_red(f"\n{traceback_str}"), file=sys.stderr)
        print(_bold(_red("\n* Internal compiler error! This is a bug; please report it.\n")), file=sys.stderr)
        sys.exit(1)


def _emit_results(
    content_bundle: external_types.ContentBundle,
    args: argparse.Namespace,
    path_to_output_file: Path | None,
    quiet: bool = False
) -> None:
    """Emits the compiled content bundle according to the user's specified output parameters.

    Args:
        content_bundle: A compiled content bundle.
        args: Parsed command-line arguments.
        path_to_output_file: Relative or absolute path to which to write an output file.
        quiet: Whether to suppress status output (errors will still be printed).

    Side Effects:
        The results are written to file and/or printed to `stderr` and `stdout`, depending on user parameters.
    """
    # If we're to print out the result, let's do so now, via `stdout` (with headers piped to `stderr`)
    if args.print is not None:
        output_filter: str = args.print.strip()
        if output_filter:
            try:
                output = _resolve_output_filter_path(content_bundle=content_bundle, path=output_filter)
            except (KeyError, IndexError, ValueError):
                error_message = f"* Invalid print filter: '{output_filter}' is not a valid path in the content bundle"
                print(_bold(_red(f"\n{error_message}\n")), file=sys.stderr)
                sys.exit(1)
        else:
            output = content_bundle
        if not quiet:
            print(_cyan(f"\n* Output{' (filtered)' if output_filter else ''}: \n"), file=sys.stderr)
        sys.stdout.write(f"{json.dumps(output, indent=2, sort_keys=True)}\n")
    # If we're to list out the compiled actions, let's do so now. We'll only write to `stderr`,
    # since the API should be used if a user wants to programmatically gather this list.
    if args.list:
        _list_compiled_constructs(content_bundle=content_bundle)
    # If an output file path has been provided, write the output file to the specified path
    if path_to_output_file:
        with open(path_to_output_file, "w", encoding="utf-8") as outfile:
            outfile.write(json.dumps(content_bundle, ensure_ascii=False))
        if not quiet:
            print(_cyan(f"\n* Wrote output to file: {path_to_output_file}\n"), file=sys.stderr)
    else:
        if not quiet:
            print(_cyan(f"\n* No output file specified\n"), file=sys.stderr)


def _resolve_output_filter_path(content_bundle: external_types.ContentBundle, path: str) -> Any:
    """Resolve a dot-notation path into a JSON-like structure.

    Args:
        content_bundle: The content bundle serving as the root structure to traverse.
        path: A dot-separated path (e.g., 'plans.foobar.phases').

    Returns:
        The value at the given path.

    Raises:
        KeyError: A segment of the path does not exist.
    """
    data: Any = dict(content_bundle)
    for segment in path.split("."):
        if isinstance(data, dict):
            data = data[segment]
        elif isinstance(data, list):
            data = data[int(segment)]
        else:
            raise KeyError(segment)
    return data


def _list_compiled_constructs(content_bundle: external_types.ContentBundle) -> None:
    """Logs all the constructs contained in the given compiled content bundle.

    Note: We only write to `stderr` here. The API should be used if a user wants
    to programmatically gather this list.

    Args:
        content_bundle: A compiled content bundle.

    Side Effects:
        The constructs are logged to `stderr`.
    """
    _print_compiled_constructs_list_block(
        block_label="Actions",
        construct_names=sorted(content_bundle["actions"])
    )
    _print_compiled_constructs_list_block(
        block_label="Action selectors",
        construct_names=sorted(content_bundle["actionSelectors"])
    )
    _print_compiled_constructs_list_block(
        block_label="Plans",
        construct_names=sorted(content_bundle["plans"])
    )
    _print_compiled_constructs_list_block(
        block_label="Plan selectors",
        construct_names=sorted(content_bundle["planSelectors"])
    )
    _print_compiled_constructs_list_block(
        block_label="Queries",
        construct_names=sorted(content_bundle["queries"])
    )
    _print_compiled_constructs_list_block(
        block_label="Sifting patterns",
        construct_names=sorted(content_bundle["siftingPatterns"])
    )
    _print_compiled_constructs_list_block(
        block_label="Tropes",
        construct_names=sorted(content_bundle["tropes"])
    )


def _print_compiled_constructs_list_block(block_label: str, construct_names: list[str]) -> None:
    """Print a block listing the given construct names.

    Args:
        block_label: A label for the block, identifying the construct type.
        construct_names: The names of the constructs to include in the block, if any, else an empty list.

    Side Effects:
        The construct block is logged to `stderr`.
    """
    if not construct_names:
        print(_dim(f"\n* {block_label} ({len(construct_names)})"), file=sys.stderr)
        return
    print(_cyan(f"\n* {block_label} ({len(construct_names)}):\n"), file=sys.stderr)
    print("\n".join(f"  - {construct_name}" for construct_name in construct_names), file=sys.stderr)


def _styled(text: str, code: str) -> str:
    """Wrap text in an ANSI escape code, respecting terminal and NO_COLOR conventions.

    Args:
        text: The text to style.
        code: An SGR parameter string.

    Returns:
        The styled text, if the environment supports styling, else the original text.
    """
    if not sys.stderr.isatty() or os.environ.get("NO_COLOR") is not None:
        return text
    return f"\033[{code}m{text}\033[0m"


def _dim(text: str) -> str:
    """Return the given string styled as dim text, if the environment supports it.

    Args:
        text: The text to style.

    Returns:
        The styled text, if the environment supports styling, else the original text.
    """
    return _styled(text, "2")


def _red(text: str) -> str:
    """Return the given string styled as red text, if the environment supports it.

    Args:
        text: The text to style.

    Returns:
        The styled text, if the environment supports styling, else the original text.
    """
    return _styled(text, "31")


def _green(text: str) -> str:
    """Return the given string styled as green text, if the environment supports it.

    Args:
        text: The text to style.

    Returns:
        The styled text, if the environment supports styling, else the original text.
    """
    return _styled(text, "32")


def _yellow(text: str) -> str:
    """Return the given string styled as yellow text, if the environment supports it.

    Args:
        text: The text to style.

    Returns:
        The styled text, if the environment supports styling, else the original text.
    """
    return _styled(text, "33")


def _cyan(text: str) -> str:
    """Return the given string styled as cyan text, if the environment supports it.

    Args:
        text: The text to style.

    Returns:
        The styled text, if the environment supports styling, else the original text.
    """
    return _styled(text, "36")


def _bold(text: str) -> str:
    """Return the given string styled as boldface text, if the environment supports it.

    Args:
        text: The text to style.

    Returns:
        The styled text, if the environment supports styling, else the original text.
    """
    return _styled(text, "1")


if __name__ == "__main__":
    main()



"""Mixin class containing visitor methods associated with Viv files."""

__all__ = ["VisitorMixinFiles"]

from typing import Any

from arpeggio import NonTerminal, PTNodeVisitor

from viv_compiler import errors, external_types, internal_types, utils


class VisitorMixinFiles(PTNodeVisitor):
    """A visitor mixin for Viv files."""

    @staticmethod
    def visit_file(node: NonTerminal, children: list[Any]) -> internal_types.AST:
        """Visit the <file> node, i.e., the root of the parse tree."""
        includes: list[str] = []
        actions: list[internal_types.IntermediateActionDefinition] = []
        action_selectors: list[internal_types.IntermediateActionSelectorDefinition] = []
        plans: list[internal_types.IntermediatePlanDefinition] = []
        plan_selectors: list[internal_types.IntermediatePlanSelectorDefinition] = []
        queries: list[internal_types.IntermediateQueryDefinition] = []
        sifting_patterns: list[internal_types.IntermediateSiftingPatternDefinition] = []
        tropes: list[internal_types.IntermediateTropeDefinition] = []
        children: list[str | dict[str, Any]]
        for unit_body in children:
            if isinstance(unit_body, str):  # Include declaration
                includes.append(unit_body)
                continue
            match unit_body['type']:
                case external_types.ConstructDiscriminator.ACTION:
                    actions.append(unit_body)
                case external_types.ConstructDiscriminator.ACTION_SELECTOR:
                    action_selectors.append(unit_body)
                case external_types.ConstructDiscriminator.PLAN:
                    plans.append(unit_body)
                case external_types.ConstructDiscriminator.PLAN_SELECTOR:
                    plan_selectors.append(unit_body)
                case external_types.ConstructDiscriminator.QUERY:
                    queries.append(unit_body)
                case external_types.ConstructDiscriminator.SIFTING_PATTERN:
                    sifting_patterns.append(unit_body)
                case external_types.ConstructDiscriminator.TROPE:
                    tropes.append(unit_body)
                case _:
                    raise errors.VivCompileError(
                        f"Encountered unexpected construct type: {unit_body['type']}",
                        source=utils.derive_source_annotations(node=node)
                    )
        ast = internal_types.AST(
            _includes=includes,
            actions=actions,
            actionSelectors=action_selectors,
            plans=plans,
            planSelectors=plan_selectors,
            queries=queries,
            siftingPatterns=sifting_patterns,
            tropes=tropes
        )
        return ast

    @staticmethod
    def visit_include(_, children: list[Any]) -> str:
        """Visit an <include> node."""
        return children[0]

    @staticmethod
    def visit_file_path(_, children: list[Any]) -> str:
        """Visit a <file_path> node."""
        return ''.join(children)



"""Mixin class containing visitor methods associated with trope definitions."""

__all__ = ["VisitorMixinTropes"]

from typing import Any

from arpeggio import NonTerminal, PTNodeVisitor

from viv_compiler import errors, external_types, internal_types, utils, validation


class VisitorMixinTropes(PTNodeVisitor):
    """A visitor mixin for Viv tropes."""

    @staticmethod
    def visit_trope(node: NonTerminal, children: list[Any]) -> internal_types.IntermediateTropeDefinition:
        """Visit a <trope> node."""
        # Populate the fields that were authored
        accumulated_fields = {}
        trope_name, trope_body = children
        if "roles" not in trope_body:
            raise errors.VivCompileError(
                f"Trope '{trope_name}' is missing 'roles' field (required)",
                source=utils.derive_source_annotations(node=node)
            )
        for field_name, field_value in trope_body.items():
            if field_name == "roles":
                validation.prevalidate_role_names(
                    construct_type=external_types.ConstructDiscriminator.TROPE,
                    construct_name=trope_name,
                    role_definitions=field_value,
                    source=utils.derive_source_annotations(node=node)
                )
                accumulated_fields["roles"] = {role["name"]: role for role in field_value}
            else:
                accumulated_fields[field_name] = field_value
        # Package up the intermediate definition
        intermediate_trope_definition = internal_types.IntermediateTropeDefinition(
            type=external_types.ConstructDiscriminator.TROPE,
            name=trope_name,
            roles=accumulated_fields["roles"],
            _conditions_raw=accumulated_fields.get("_conditions_raw", []),
        )
        return intermediate_trope_definition

    @staticmethod
    def visit_trope_header(_, children: list[external_types.TropeName]) -> external_types.TropeName:
        """Visit a <trope_header> node."""
        return children[0]

    @staticmethod
    def visit_trope_body(_, children: list[Any]) -> dict[str, Any]:
        """Visit a <trope_body> node."""
        accumulated_fields = {}
        for child in children:
            accumulated_fields.update(child)
        return accumulated_fields

    @staticmethod
    def visit_trope_roles(
        _, children: list[external_types.RoleDefinition]
    ) -> dict[str, list[external_types.RoleDefinition]]:
        """Visit a <trope_roles> node."""
        return {"roles": children}

    @staticmethod
    def visit_trope_conditions(
        _, children: list[list[external_types.Expression]]
    ) -> dict[str, list[external_types.Expression]]:
        """Visit a <trope_conditions> node."""
        return {"_conditions_raw": children[0]}



"""System configuration for the Viv DSL compiler.

Important: These settings are only meant to be modified in the course of system development, i.e.,
by the compiler maintainers and not by end users. That said, certain user-supplied default values
will be used programmatically to update some of the default values below.
"""

from typing import Final

from viv_compiler import external_types


# A default importance value, to be used when the CLI user (or API caller) does not provide one
DEFAULT_IMPORTANCE_SCORE = 1.0

# A default salience value, to be used when the CLI user (or API caller) does not provide one
DEFAULT_SALIENCE_SCORE = 1.0

# A default reaction priority value, to be used when the CLI user (or API caller) does not provide one
DEFAULT_REACTION_PRIORITY_VALUE = 1.0

# Name of the root symbol in the Viv DSL grammar
GRAMMAR_ROOT_SYMBOL: Final = "file"

# Name of the symbol associated with comments in the Viv DSL grammar
GRAMMAR_COMMENT_SYMBOL: Final = "comment"

# Viv expression types that support negation. While the grammar (and therefore the parser)
# will allow for negation in other expression types, the validator will use this config
# parameter to enforce this policy.
NEGATABLE_EXPRESSION_TYPES: Final = {
    external_types.ExpressionDiscriminator.ACTION_RELATION,
    external_types.ExpressionDiscriminator.CUSTOM_FUNCTION_CALL,
    external_types.ExpressionDiscriminator.COMPARISON,
    external_types.ExpressionDiscriminator.CONJUNCTION,
    external_types.ExpressionDiscriminator.DISJUNCTION,
    external_types.ExpressionDiscriminator.ENTITY_REFERENCE,
    external_types.ExpressionDiscriminator.LOOP,
    external_types.ExpressionDiscriminator.MEMBERSHIP_TEST,
    external_types.ExpressionDiscriminator.MEMORY_CHECK,
    external_types.ExpressionDiscriminator.SYMBOL_REFERENCE,
    external_types.ExpressionDiscriminator.TROPE_FIT,
}

# A tuple containing the construct types that take on initiator roles
CONSTRUCT_TYPES_USING_INITIATOR_ROLES: Final = (
    external_types.ConstructDiscriminator.ACTION,
    external_types.ConstructDiscriminator.ACTION_SELECTOR,
)

# Mapping from construct type to the role labels permitted in construct definitions of that type
ROLE_LABELS_PERMITTED_IN_CONSTRUCT_OF_TYPE: Final = {
    external_types.ConstructDiscriminator.ACTION: (
        'action', 'anywhere', 'bystander', 'character', 'initiator', 'item',
        'location', 'partner', 'precast', 'recipient', 'spawn', 'symbol'
    ),
    external_types.ConstructDiscriminator.ACTION_SELECTOR: (
        'action', 'anywhere', 'character', 'item', 'location', 'precast', 'symbol', 'initiator'
    ),
    external_types.ConstructDiscriminator.PLAN: (
        'action', 'anywhere', 'character', 'item', 'location', 'precast', 'symbol'
    ),
    external_types.ConstructDiscriminator.PLAN_SELECTOR: (
        'action', 'anywhere', 'character', 'item', 'location', 'precast', 'symbol'
    ),
    external_types.ConstructDiscriminator.QUERY: (
        'action', 'anywhere', 'character', 'item', 'location', 'precast', 'symbol'
    ),
    external_types.ConstructDiscriminator.SIFTING_PATTERN: (
        'action', 'anywhere', 'character', 'item', 'location', 'precast', 'symbol'
    ),
    external_types.ConstructDiscriminator.TROPE: (
        'action', 'anywhere', 'character', 'item', 'location', 'precast', 'symbol'
    ),
}

# Sets of role labels such that only one from the set may be present for a given role
MUTUALLY_INCOMPATIBLE_ROLE_LABELS: Final = (
    ("character", "item", "location", "action", "symbol"),  # Not possible, but belt and suspenders
    ("initiator", "partner", "recipient", "bystander", "anywhere"),
    ("spawn", "symbol"),
    ("spawn", "action"),
    ("spawn", "initiator"),
)

# Name for the special action self-reference role, which is always bound to the action itself
ACTION_SELF_REFERENCE_ROLE_NAME: Final = "this"

# A set containing the special role names that are automatically created by Viv at various points
SPECIAL_ROLE_NAMES: Final = {"hearer", ACTION_SELF_REFERENCE_ROLE_NAME}

# The path to which the scratch-variable sigil `$` expands. This sigil is really just syntactic sugar for
# the path `@this.scratch`, which stores a blackboard local to a performed action. For instance, the scratch
# operation `$@foo.bar = 99` is syntactic sugar for the expression `@this.scratch.foo.bar = 99`.
SCRATCH_VARIABLE_REFERENCE_ANCHOR: Final = ACTION_SELF_REFERENCE_ROLE_NAME
SCRATCH_VARIABLE_REFERENCE_PATH_PREFIX: Final = [
    external_types.ReferencePathComponentPropertyName(
        type=external_types.ReferencePathComponentDiscriminator.REFERENCE_PATH_COMPONENT_PROPERTY_NAME,
        name="scratch",
        failSafe=False
    )
]

# A tuple containing the names of action fields in which assignments are permitted
ACTION_FIELDS_PERMITTING_ASSIGNMENTS: Final = ("scratch", "effects",)

# A tuple containing the names of action fields in which reactions are permitted
ACTION_FIELDS_PERMITTING_REACTIONS: Final = ("reactions",)

# A tuple containing discriminators for the construct types in which reactions are permitted
CONSTRUCT_TYPES_PERMITTING_REACTIONS: Final = (
    external_types.ConstructDiscriminator.ACTION,
    external_types.ConstructDiscriminator.PLAN,
)

# A tuple containing the names of query fields defined using set predicates
QUERY_FIELDS_TAKING_SET_PREDICATES: Final = (
    "tags", "associations", "location", "initiator", "partners",
    "recipients", "bystanders", "active", "present",
)

# A tuple containing the names of query fields defined using set predicates that do not allow
# the operator 'all'. These fields pertain to singular values, and as a result, the 'all'
# operator does not make sense for them.
QUERY_FIELDS_WITH_SINGULAR_VALUE: Final = ("location", "initiator",)

# The role name to use for a selector virtual initiator role, which is created when an author
# defines a selector that elides the `roles` field. This supports the "initiator pass-through"
# pattern for selectors, where targeting a selector means targeting the selector candidates
# with the same initiator who is at hand (during targeting of the selector).
SELECTOR_VIRTUAL_INITIATOR_ROLE_NAME: Final = "__initiator__"

# A transient prefix used for positional role bindings, resolved to the corresponding
# role name during postprocessing.
POSITIONAL_BINDING_TRANSIENT_PREFIX: Final = "__position__"



"""Internal functionality for validating query definitions."""

__all__ = ["validate_query_definitions"]

from viv_compiler import config, errors, external_types
from ._common import validate_common_concerns


def validate_query_definitions(content_bundle: external_types.ContentBundle) -> None:
    """Validate the query definitions in the given compiled content bundle.

    Args:
        content_bundle: A compiled content bundle.

    Returns:
        None. Only returns if no issue is encountered.
    """
    for query_definition in content_bundle["queries"].values():
        # Validate general issues that pertain to all construct types
        validate_common_concerns(construct_definition=query_definition, content_bundle=content_bundle)
        # Validate `actionName` field
        _validate_query_action_field(query_definition=query_definition, content_bundle=content_bundle)
        # Validate `time` field
        _validate_query_time_field(query_definition=query_definition)
        # Validate fields defined using set predicates
        _validate_query_set_predicates(query_definition=query_definition)


def _validate_query_action_field(
    query_definition: external_types.QueryDefinition,
    content_bundle: external_types.ContentBundle
) -> None:
    """Validate the given query definition's `actionName` field.

    Args:
        query_definition: A query definition in a compiled content bundle.
        content_bundle: A compiled content bundle.

    Returns:
        None. Only returns if no issue is encountered.

    Raises:
        VivCompileError: The query definition has an invalid `actionName` field.
    """
    action_name_field = query_definition["actionName"]
    if action_name_field is None:
        return
    for set_predicate in action_name_field:
        for action_name_expression in set_predicate["operand"]:
            action_name = action_name_expression["value"]
            if action_name not in content_bundle["actions"]:
                raise errors.VivCompileError(
                    msg=f"Query '{query_definition['name']}' references undefined action: '{action_name}'",
                    source=action_name_expression['source'],
                )


def _validate_query_time_field(query_definition: external_types.QueryDefinition) -> None:
    """Validate the given query definition's `time` field.

    Args:
        query_definition: A query definition in a compiled content bundle.

    Returns:
        None. Only returns if no issue is encountered.

    Raises:
        VivCompileError: The query definition has an invalid `time` field.
    """
    time_field = query_definition['time']
    if time_field is None:
        return
    time_of_day_constraints = 0
    time_frame_constraints = 0
    for temporal_constraint in time_field:
        if temporal_constraint['type'] == external_types.TemporalStatementDiscriminator.TIME_OF_DAY:
            time_of_day_constraints += 1
            if time_of_day_constraints > 1:
                raise errors.VivCompileError(
                    f"Query '{query_definition['name']}' has multiple time-of-day "
                    "temporal constraints (only one is allowed)"
                )
        else:
            time_frame_constraints += 1
            if time_frame_constraints > 1:
                raise errors.VivCompileError(
                    f"Query '{query_definition['name']}' has multiple time-frame temporal "
                    f"constraints (only one is allowed)"
                )


def _validate_query_set_predicates(query_definition: external_types.QueryDefinition) -> None:
    """Validate the use of set predicates in the given query definition.

    Args:
        query_definition: A query definition in a compiled content bundle.

    Returns:
        None. Only returns if no issue is encountered.

    Raises:
        VivCompileError: The query definition makes invalid use of set predicates.
    """
    for field_name, potential_set_predicates in query_definition.items():
        if field_name not in config.QUERY_FIELDS_TAKING_SET_PREDICATES:
            continue
        if potential_set_predicates is None:
            continue
        set_predicates: list[external_types.SetPredicate] = potential_set_predicates
        for set_predicate in set_predicates:
            if set_predicate["operator"] == "all":
                if field_name in config.QUERY_FIELDS_WITH_SINGULAR_VALUE:
                    raise errors.VivCompileError(
                        f"Query '{query_definition['name']}' uses 'all' operator in '{field_name}' field, "
                        f"but this is not allowed because this field corresponds to a singular value "
                        f"(use 'exactly' to specify a single possibility)"
                    )
            for operand_component in set_predicate['operand']:
                if set_predicate['operand'].count(operand_component) > 1:
                    raise errors.VivCompileError(
                        f"Query '{query_definition['name']}' has duplicate value following "
                        f"'{set_predicate['operator']}' operator in the '{field_name}' field"
                    )
            if set_predicate["operator"] == "exactly":
                if field_name in config.QUERY_FIELDS_WITH_SINGULAR_VALUE:
                    if len(set_predicate['operand']) > 1:
                        raise errors.VivCompileError(
                            f"Query '{query_definition['name']}' uses 'exactly' operator in '{field_name}' field, "
                            f"but then has multiple operand components (only a single '{field_name}' can be "
                            f"specified when 'exactly' is used)"
                        )
                if len(set_predicates) > 1:
                    raise errors.VivCompileError(
                        f"Query '{query_definition['name']}' uses 'exactly' operator in '{field_name}' field, "
                        f"but then has additional criteria ('exactly' must be alone when present)"
                    )
